@@ -27,6 +27,7 @@ import sys
 import os
 import asyncio
 import typing as tp
+import functools
 try:
     import aioprocessing
 except ImportError:
@@ -186,14 +187,14 @@ class ProcManager(object):
                 try:
                     task = self.tasks.get(False)
                 except EmptyQueue:
-                    self.logger.error("Queue is empty! Take a poison pill!")
+                    self.logger.warning("Queue is empty! Take a poison pill!")
                     task = None
                 if task is None:  # poison pill
                     self.logger.debug("Got poison pill")
                     break
                 self.logger.debug("Got task, mark as defer")
                 mark_defer_counter += 1
-                sock, handler_modulename, handler_classname = uncompress_task(task)
+                sock, handler_modulename, handler_classname, port = uncompress_task(task)
                 handler_class = getattr(importlib.import_module(handler_modulename), handler_classname)
                 handler_instance = handler_class(sock, self.config)
                 handler_instance.defer(return_message)
@@ -218,13 +219,13 @@ class ProcManager(object):
             remaining_timeout = max(join_timeout - tpassed, 0.05)
             worker.join(int(remaining_timeout))
             if worker.is_alive():
-                self.logger.error("Could not stop worker %s (pid: %u) with given timeout of %f (%f)"
-                                  % (worker, worker.pid, join_timeout, remaining_timeout))
+                self.logger.warning("Could not stop worker %s (pid: %u) with given timeout of %f (%f)"
+                                    % (worker, worker.pid, join_timeout, remaining_timeout))
                 worker.terminate()
                 time.sleep(0.1)
                 if worker.is_alive():
-                    self.logger.error("Could not stop worker %s (pid: %u) with SIGTERM, use SIGKILL"
-                                      % (worker, worker.pid))
+                    self.logger.warning("Could not stop worker %s (pid: %u) with SIGTERM, use SIGKILL"
+                                        % (worker, worker.pid))
                     try:
                         worker.kill()
                     except AttributeError:
@@ -311,11 +312,21 @@ def fuglu_async_process_worker(queue, config, shared_state, child_to_server_mess
     if not async_threadpoolworkers:
         async_threadpoolworkers = max(1, int(async_coroutines/2))
 
+    try:
+        async_prepare_timeout = config.getfloat('performance', 'async_prepare_timeout')
+    except Exception as e:
+        import traceback
+        traceback.print_exc()
+        async_prepare_timeout = None
+
     workerstate = WorkerStateWrapper(shared_state,'loading configuration', async_coroutines=async_coroutines)
     logger = logging.getLogger('fuglu.process.%s(%u)' % (workerstate.process.name, workerstate.process.pid))
     logger.debug(f"New worker: {logtools.createPIDinfo()}, "
                  f"coroutines={async_coroutines} with threadpool{async_threadpoolworkers}")
-
+    logger.debug(f"async timeouts: "
+                 f"coroutines={async_coroutines}, "
+                 f"threadworkers={async_threadpoolworkers}, "
+                 f"preparetimeout={async_prepare_timeout}")
 
     # Setup address compliance checker
     # -> Due to default linux forking behavior this should already
@@ -383,7 +394,7 @@ def fuglu_async_process_worker(queue, config, shared_state, child_to_server_mess
 
         with ThreadPoolExecutor(max_workers=async_threadpoolworkers) as pool:
             # create worker coroutines for async loop
-            asworkers = [fuglu_async_process_worker_main(queue, workerstate, logger, controller, i, logtimings, pool) for i in range(1, 1+async_coroutines)]
+            asworkers = [fuglu_async_process_worker_main(queue, workerstate, logger, controller, i, logtimings, pool, async_prepare_timeout) for i in range(1, 1+async_coroutines)]
             logger.debug(f"We put {len(asworkers)} into the loop...")
             loop.run_until_complete(asyncio.gather(*asworkers, loop=loop))
             logger.debug("Gathered result...")
@@ -420,8 +431,49 @@ def fuglu_async_process_worker(queue, config, shared_state, child_to_server_mess
         logger.debug("%s: async proc worker finished & complete..." % logtools.createPIDinfo())
 
 
-async def fuglu_async_process_worker_main(queue, workerstate, logger, controller, asyncid, logtimings, pool):
+def prepare_session_task_reader(asyncid: int, workerstate, logger, task, loop):
+    """First part for preparing session: uncompress task & create async reader"""
+    workerstate.set_workerstate('starting scan session', id=asyncid)
+    logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Child process starting scan session")
+
+    sock, handler_modulename, handler_classname, port = uncompress_task(task)
+
+    assert sm.MilterHandler.__name__ == handler_classname, \
+        f"{sm.MilterHandler.__name__} != {handler_classname} => asyncprocpool only works for milter {sm.MilterHandler.__name__} != {handler_classname}!!!"
+
+    reader = asyncio.StreamReader(loop=loop)
+    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
+    return sock, handler_modulename, handler_classname, port, reader, protocol
+
+
+def prepare_session_writer_handler(asyncid: int, workerstate, logger, controller,
+                                   logtimings: bool, pool, transport, protocol, reader, loop, port):
+    """Second part for preparing session: create async writer and Milterhandler/-session"""
+
+    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
+
+    logger.debug(
+        f"{logtools.createPIDinfo()}({asyncid}): Extracted reader/writer -> create MilterHandler & MilterSession")
+    # create milter handler
+    mhand = asm.MilterHandler(controller.config,
+                              controller.prependers,
+                              controller.plugins,
+                              controller.appenders,
+                              port,
+                              controller.milterdict,
+                              workerstate=workerstate,
+                              asyncid=asyncid,
+                              enable=logtimings,
+                              pool=pool)
+
+    # create milter session, passing handler
+    msess = asm.MilterSession(reader, writer, controller.config, options=mhand.sess_options, mhandler=mhand)
+    return writer, mhand, msess
+
+
+async def fuglu_async_process_worker_main(queue, workerstate, logger, controller, asyncid, logtimings, pool, async_prepare_timeout):
     # define queue type here to prevent an error if aioprocessing module is not present
+    logger.info(f"{logtools.createPIDinfo()}({asyncid}): Child asworker process ready")
     queue: aioprocessing.AioQueue
     while True:
         try:
@@ -442,43 +494,76 @@ async def fuglu_async_process_worker_main(queue, workerstate, logger, controller
                     pass
                 finally:
                     return
-            workerstate.set_workerstate('starting scan session', id=asyncid)
-            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Child process starting scan session")
-
-            sock, handler_modulename, handler_classname, port = uncompress_task(task)
-
-            assert sm.MilterHandler.__name__ == handler_classname, f"{sm.MilterHandler.__name__} != {handler_classname} => asyncprocpool only works for milter!!!"
 
             loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
+            if pool:
+                logger.debug(f"Prepare session (reader, timeout={async_prepare_timeout}) using executor from pool")
+                future = loop.run_in_executor(pool,
+                                              functools.partial(prepare_session_task_reader,
+                                                                asyncid=asyncid,
+                                                                workerstate=workerstate,
+                                                                logger=logger,
+                                                                task=task,
+                                                                loop=loop
+                                                                )
+                                              )
+                if async_prepare_timeout:
+                    sock, handler_modulename, handler_classname, port, reader, protocol \
+                        = await asyncio.wait_for(future, timeout=async_prepare_timeout)
+                else:
+                    sock, handler_modulename, handler_classname, port, reader, protocol \
+                        = await future
+            else:
+                logger.debug("Prepare session (reader) directly")
+                sock, handler_modulename, handler_classname, port, reader, protocol \
+                    = prepare_session_task_reader(asyncid, workerstate, logger, task, loop)
 
-            reader = asyncio.StreamReader(loop=loop)
-            protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
+            future = loop.create_connection(lambda: protocol, sock=sock)
+
+            logger.debug(f"Prepare session (connection, timeout={async_prepare_timeout})")
+            if async_prepare_timeout:
+                transport, _ = await asyncio.wait_for(future , timeout=async_prepare_timeout)
+            else:
+                transport, _ = await future
 
-            transport, _ = await loop.create_connection(lambda: protocol, sock=sock)
             logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Extracted transport & protocol")
 
-            writer = asyncio.StreamWriter(transport, protocol, reader, loop)
+            if pool:
+                logger.debug(f"Prepare session (objects, timeout={async_prepare_timeout}) using executor from pool")
+                future = loop.run_in_executor(pool,
+                                              functools.partial(prepare_session_writer_handler,
+                                                                asyncid=asyncid,
+                                                                workerstate=workerstate,
+                                                                logger=logger,
+                                                                controller=controller,
+                                                                logtimings=logtimings,
+                                                                pool=pool,
+                                                                transport=transport,
+                                                                protocol=protocol,
+                                                                reader=reader,
+                                                                loop=loop,
+                                                                port=port
+                                                                )
+                                              )
+
+                if async_prepare_timeout:
+                    writer, mhand, msess = await asyncio.wait_for(future, timeout=async_prepare_timeout)
+                else:
+                    writer, mhand, msess = await future
+            else:
+                logger.debug("Prepare session (objects) directly")
+                writer, mhand, msess = prepare_session_writer_handler(asyncid=asyncid, workerstate=workerstate,
+                                                                      logger=logger, controller=controller,
+                                                                      logtimings=logtimings, pool=pool,
+                                                                      transport=transport, protocol=protocol,
+                                                                      reader=reader, loop=loop, port=port)
 
-            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Extracted reader/writer -> create MilterHandler & MilterSession")
-            # create milter handler
-            mhand = asm.MilterHandler(controller.config,
-                                      controller.prependers,
-                                      controller.plugins,
-                                      controller.appenders,
-                                      port,
-                                      controller.milterdict,
-                                      workerstate=workerstate,
-                                      asyncid=asyncid,
-                                      enable=logtimings,
-                                      pool=pool)
 
-            # create milter session, passing handler
-            msess = asm.MilterSession(reader, writer, controller.config,
-                                      options=mhand.sess_options, mhandler=mhand)
             # handle session
-            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Handle session...")
+            workerstate.set_workerstate(f'{msess.id} Handle scan session...', id=asyncid)
+            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Handle scan session {msess.id}")
             await msess.handlesession()
-            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Finished handling session...")
+            logger.debug(f"{logtools.createPIDinfo()}({asyncid}): Finished handling session {msess.id}...")
 
             del reader
             del protocol
@@ -494,6 +579,30 @@ async def fuglu_async_process_worker_main(queue, workerstate, logger, controller
         except KeyboardInterrupt:
             logger.error(f"{logtools.createPIDinfo()}({asyncid}): Got Keyboard Interrupt -> leaving while - loop")
             return
+        except OSError as e:
+            logger.warning(f"{logtools.createPIDinfo()}({asyncid}): Got OSError {str(e)}, continue...")
+            #
+            # This error happens from time to time, not sure why, maybe if connection gets killed on the outside
+            # by postfix. Traceback example:
+            #
+            # Traceback (most recent call last):
+            #   File "/fuglu/fuglu/asyncprocpool.py", line 448, in fuglu_async_process_worker_main
+            #     sock, handler_modulename, handler_classname, port = uncompress_task(task)
+            #   File "/fuglu/fuglu/protocolbase.py", line 220, in uncompress_task
+            #     sock = pickle.loads(pickled_socket)
+            #   File "/usr/local/lib/python3.9/multiprocessing/reduction.py", line 246, in _rebuild_socket
+            #     fd = df.detach()
+            #   File "/usr/local/lib/python3.9/multiprocessing/resource_sharer.py", line 58, in detach
+            #     return reduction.recv_handle(conn)
+            #   File "/usr/local/lib/python3.9/multiprocessing/reduction.py", line 188, in recv_handle
+            #     with socket.fromfd(conn.fileno(), socket.AF_UNIX, socket.SOCK_STREAM) as s:
+            #   File "/usr/local/lib/python3.9/socket.py", line 545, in fromfd
+            #     return socket(family, type, proto, nfd)
+            #   File "/usr/local/lib/python3.9/socket.py", line 232, in __init__
+            #     _socket.socket.__init__(self, family, type, proto, fileno)
+            # OSError: [Errno 9] Bad file descriptor
+        except asyncio.TimeoutError as e:
+            logger.error(f"{logtools.createPIDinfo()}({asyncid}): Async-Timeout {str(e)}", exc_info=e)
         except Exception as e:
             logger.error(f"{logtools.createPIDinfo()}({asyncid}): Exception {str(e)}", exc_info=e)
 
