@@ -22,6 +22,7 @@ import os
 from fuglu.protocolbase import ProtocolHandler, BasicTCPServer
 from fuglu.shared import Suspect, apply_template
 from fuglu.stringencode import force_bString, force_uString
+from fuglu.bounce import FugluSMTPClient
 from .smtpconnector import smtp_strip_address
 
 from email.header import Header
@@ -75,7 +76,8 @@ class ESMTPHandler(ProtocolHandler):
 
         code, answer = self.sess.forwardconn.data(force_bString(msgcontent))
         answer = force_uString(answer)
-        return code, answer
+        queueid = self.sess.forwardconn.queueid
+        return code, answer, queueid
 
     def get_suspect(self, **kwargs):
         success = self.sess.getincomingmail()
@@ -104,8 +106,9 @@ class ESMTPHandler(ProtocolHandler):
         return suspect
 
     def commitback(self, suspect):
-        injectcode, injectanswer = self.re_inject(suspect)
+        injectcode, injectanswer, queueid = self.re_inject(suspect)
         suspect.set_tag("injectanswer", injectanswer)
+        suspect.set_tag("injectqueueid", queueid)
 
         values = dict(injectanswer=injectanswer)
         message = apply_template(self.config.get('esmtpconnector', 'queuetemplate'), suspect, values)
@@ -252,7 +255,14 @@ class ESMTPPassthroughSession(object):
             targethost = self.config.get('main', 'outgoinghost')
             if targethost == '${injecthost}':
                 targethost = self.socket.getpeername()[0]
-            self.forwardconn = smtplib.SMTP(force_uString(targethost), self.config.getint('main', 'outgoingport'))
+            elif targethost.startswith("$"):
+                # extract target host from environment variable
+                env_targethost = os.environ[targethost[1:]]
+                if not env_targethost:
+                    raise ValueError(f"Could not extract target host from environment var '{targethost}'")
+                targethost = env_targethost
+
+            self.forwardconn = FugluSMTPClient(force_uString(targethost), self.config.getint('main', 'outgoingport'))
         self.logger.debug("""SEND: "%s" """ % command)
 
         # docmd seems to have a normal string as input, so
