@@ -19,6 +19,7 @@ import logging
 import socket
 import tempfile
 import os
+import time
 import typing as tp
 
 from typing import Tuple, Union
@@ -165,8 +166,8 @@ class MilterHandler(ProtocolHandler):
         self.logger.debug("After getting incoming mail...")
 
         sess = self.sess
-        from_address = sess.get_cleaned_from_address()
-        recipients = sess.get_cleaned_recipients()
+        from_address = force_uString(sess.get_cleaned_from_address())
+        recipients = force_uString(sess.get_cleaned_recipients())
 
         # If there's no file
         temp_filename = sess.tempfilename
@@ -176,17 +177,18 @@ class MilterHandler(ProtocolHandler):
         # If there is a filename but no file
         if temp_filename and not os.path.exists(temp_filename):
             self.logger.warning("File '%s' not found for suspect creation! from: %s, to: %s"
-                                % (temp_filename, str(from_address), str(recipients)))
+                                % (temp_filename, from_address, recipients))
             return None
 
         suspect = Suspect(from_address, recipients, temp_filename, att_cachelimit=self._att_mgr_cachesize,
                           att_defaultlimit=self._att_defaultlimit, att_maxlimit=self._att_maxlimit,
                           sasl_login=sess.sasl_login, sasl_sender=sess.sasl_sender, sasl_method=sess.sasl_method,
                           queue_id=sess.queueid, id=id)
-
+        suspect.timestamp = sess.timestamp
+        
         # add headers
         for hdrname, hdrval in self.sess.addheaders.items():
-            suspect.addheader(key=hdrname, value=hdrval, immediate=False)
+            suspect.add_header(key=hdrname, value=hdrval, immediate=False)
 
         # add session tags to Suspect
         if self.sess.tags:
@@ -196,7 +198,7 @@ class MilterHandler(ProtocolHandler):
             '"%s" "%s"' % (suspect.id, sess.queueid if sess.queueid else "NOQUEUE"))
 
         if sess.heloname is not None and sess.addr is not None and sess.fcrdns is not None:
-            suspect.clientinfo = sess.heloname, sess.addr, sess.fcrdns
+            suspect.clientinfo = force_uString(sess.heloname), force_uString(sess.addr), force_uString(sess.fcrdns)
 
         return suspect
 
@@ -545,6 +547,7 @@ class MilterSession(lm.MilterProtocol):
                  ):
         # enable options for version 2 protocol
         super().__init__(opts=options)
+        self.timestamp = time.time()
         self.transport = sock
         try:
             self.tmpdir = config.get('main', 'tempdir')
@@ -587,6 +590,10 @@ class MilterSession(lm.MilterProtocol):
         # tags (will be passed to Suspect)
         self.tags = {}
 
+    @property
+    def sasl_user(self):
+        return self.sasl_login
+
     def add_plugin_skip(self, pluginname: str, tag: str = "skipmplugins"):
         """Add plugin to skiplist"""
         if tag in self.tags:
@@ -633,6 +640,8 @@ class MilterSession(lm.MilterProtocol):
             templdict['queueid'] = force_uString(self.queueid)
         if self.sasl_login is not None:
             templdict['sasl_login'] = force_uString(self.sasl_login)
+        if self.sasl_user is not None: # same as sasl_login
+            templdict['sasl_user'] = force_uString(self.sasl_user)
         if self.sasl_sender is not None:
             templdict['sasl_sender'] = force_uString(self.sasl_sender)
         if self.from_address:
@@ -646,6 +655,13 @@ class MilterSession(lm.MilterProtocol):
             templdict['to_domain'] = self.to_domain
         if self.size is not None:
             templdict['size'] = force_uString(self.size)
+
+        # put tags also in template dict with prefix "tag_" so it can be
+        # used in reject messages
+        if self.tags:
+            for tkey, tval in self.tags.items():
+                templdict[f'tag_{tkey}'] = f"{tval}"
+
         return templdict
 
     def reset_connection(self):
@@ -733,7 +749,7 @@ class MilterSession(lm.MilterProtocol):
 
     @staticmethod
     def extract_domain(address: str, lowercase=True):
-        if not address:
+        if not address or address.lower() == "postmaster":
             return None
         else:
             try:
@@ -921,8 +937,8 @@ class MilterSession(lm.MilterProtocol):
                                                               hostname, family, str(command_dict)))
         self.store_info_from_dict(command_dict)
         if family not in (b'4', b'6'):  # we don't handle unix socket
-            self.logger.error('Return temporary fail since family is: %s' % force_uString(family))
-            self.logger.error(u'command dict is: %s' % MilterSession.dict_unicode(command_dict))
+            self.logger.warning('Return temporary fail since family is: %s' % force_uString(family))
+            self.logger.warning(u'command dict is: %s' % MilterSession.dict_unicode(command_dict))
             return lm.TEMPFAIL
         if hostname is None or force_uString(hostname) == u'[%s]' % force_uString(ip):
             hostname = u'unknown'
@@ -938,7 +954,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_connect(sess=self, host=self.fcrdns, addr=self.addr)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {CONNECT}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {CONNECT}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
 
         return lm.CONTINUE
@@ -954,7 +970,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_helo(sess=self, helo=self.heloname)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {HELO}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {HELO}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
         return lm.CONTINUE
 
@@ -972,7 +988,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_mailfrom(sess=self, sender=self.sender)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {MAILFROM}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {MAILFROM}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
         return lm.CONTINUE
 
@@ -991,7 +1007,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_rcpt(sess=self, recipient=recipient)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {RCPT}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {RCPT}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
 
         return lm.CONTINUE
@@ -1010,7 +1026,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_header(sess=self, key=key, value=val)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {HEADER}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {HEADER}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
 
         return lm.CONTINUE
@@ -1027,7 +1043,7 @@ class MilterSession(lm.MilterProtocol):
             res = plug.examine_eoh(sess=self)
             res, msg, retcode = self.handle_milter_plugin_reply(res)
             if not res == lm.CONTINUE:
-                self.logger.info(f"{self.id} {EOH}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,'unknown')}")
+                self.logger.info(f"{self.id} {EOH}-Plugin {plug} returns non-continue-result: {RETCODE2STR.get(res,f'unknown(orig:{res})')}")
                 return retcode
         return lm.CONTINUE
 
