@@ -37,7 +37,7 @@ from fuglu.connectors.ncconnector import NCServer
 from fuglu.connectors.smtpconnector import SMTPServer
 from fuglu.debug import ControlServer, CrashStore
 from fuglu.funkyconsole import FunkyConsole
-from fuglu.shared import (HAVE_BEAUTIFULSOUP, Suspect, default_template_values)
+from fuglu.shared import HAVE_BEAUTIFULSOUP, Suspect, default_template_values
 from fuglu.mshared import BasicMilterPlugin
 from fuglu.stats import StatsThread
 from fuglu.stringencode import force_bString, force_uString
@@ -135,7 +135,7 @@ def check_version_status(lint=False):
         code = int(ret.groupdict()['replycode'])
         for bitmask, message in bitmaskmap.items():
             if code & bitmask == bitmask:
-                logging.warn(message)
+                logging.warning(message)
                 if lint:
                     fc = FunkyConsole()
                     print(fc.strcolor(message, "yellow"))
@@ -404,6 +404,12 @@ class MainController(DefConfigMixin):
                 'section': 'performance',
                 'description': "Upper limit for maximum number of files to be extracted from archives (default: 500)"
             },
+            
+            'disable_aiosmtp': {
+                'default': "False",
+                'section': 'performance',
+                'description': "do not use aiosmtplib even if it is installed (fallback to python built in smtplib)"
+            },
 
             # spam section
             'defaultlowspamaction': {
@@ -451,13 +457,36 @@ class MainController(DefConfigMixin):
                 'description': """read runtime configuration values from a database. requires sqlalchemy to be installed""",
                 'confidential': True,
             },
-
+            
             'sql': {
-                'default': """SELECT value FROM fugluconfig WHERE `section`=:section AND `option`=:option AND `scope` IN ('$GLOBAL',CONCAT('%',:to_domain),:to_address) ORDER BY `scope` DESC""",
+                'default': """SELECT value FROM fugluconfig WHERE `section`=:section AND `scope` IN ('$GLOBAL',CONCAT('%',:to_domain),:to_address) AND `option`=:option ORDER BY `scope` DESC LIMIT 1""",
                 'section': 'databaseconfig',
                 'description': """sql query that returns a configuration value override. sql placeholders are ':section',':option' in addition the usual suspect filter default values like ':to_domain', ':to_address' etc\nif the statement returns more than one row/value only the first value in the first row is used""",
             },
-
+            
+            'sqlsection': {
+                'default': """SELECT value, option, scope FROM fugluconfig WHERE `section`=:section AND `scope` IN ('$GLOBAL',CONCAT('%',:to_domain),:to_address) ORDER BY `scope` DESC""",
+                'section': 'databaseconfig',
+                'description': """sql query that returns all values, options and scopes (relevant for one recipient) of one section. sql placeholders ':section' available in addition the usual suspect filter default values like ':to_domain', ':to_address' etc""",
+            },
+    
+            'option_field': {
+                'default': "option",
+                'section': 'databaseconfig',
+                'description': """what is the name of the "option" field in the database""",
+            },
+            'scope_field': {
+                'default': "scope",
+                'section': 'databaseconfig',
+                'description': """what is the name of the "scope" field in the database""",
+            },
+            'value_field': {
+                'default': "value",
+                'section': 'databaseconfig',
+                'description': """what is the name of the "value" field in the database""",
+            },
+            
+            
             # environment
             'boundarydistance': {
                 'default': "0",
@@ -517,12 +546,17 @@ class MainController(DefConfigMixin):
             },
 
             'actionoverride': {
-                'default': "fuglu.plugins.actionoverride.ActionOverridePlugin",
+                'default': "fuglu.plugins.decision.ActionOverridePlugin",
+                'section': 'PluginAlias',
+            },
+
+            'killer': {
+                'default': "fuglu.plugins.decision.KillerPlugin",
                 'section': 'PluginAlias',
             },
 
             'icap': {
-                'default': "fuglu.plugins.icap.ICAPPlugin",
+                'default': "fuglu.plugins.antivirus.ICAPPlugin",
                 'section': 'PluginAlias',
             },
 
@@ -532,7 +566,7 @@ class MainController(DefConfigMixin):
             },
 
             'fprot': {
-                'default': "fuglu.plugins.fprot.FprotPlugin",
+                'default': "fuglu.plugins.antivirus.FprotPlugin",
                 'section': 'PluginAlias',
             },
 
@@ -618,22 +652,19 @@ class MainController(DefConfigMixin):
     def start_connector(self, portspec):
         protocol = 'smtp'
         bindaddress = self.config.get('main', 'bindaddress')
-
         protocol, bindaddress, port = MainController.get_connectorinfo(portspec, protocol, bindaddress)
 
         self.logger.info("starting connector %s/%s/%s" % (protocol, bindaddress, port))
         try:
             port = int(port)
             if protocol == 'smtp':
-                smtpserver = SMTPServer(
-                    self, port=port, address=bindaddress)
+                smtpserver = SMTPServer(self, port=port, address=bindaddress)
                 tr = threading.Thread(target=smtpserver.serve, args=())
                 tr.daemon = True
                 tr.start()
                 self.servers.append(smtpserver)
             elif protocol == 'esmtp':
-                esmtpserver = ESMTPServer(
-                    self, port=port, address=bindaddress)
+                esmtpserver = ESMTPServer(self, port=port, address=bindaddress)
                 tr = threading.Thread(target=esmtpserver.serve, args=())
                 tr.daemon = True
                 tr.start()
@@ -643,31 +674,25 @@ class MainController(DefConfigMixin):
                 backend = self.config.get('performance', 'backend')
                 if backend == "asyncio":
                     self.logger.debug("Starting ASYNCMilterServer....")
-                    milterserver = ASYNCMilterServer(
-                        self, port=port, address=bindaddress)
+                    milterserver = ASYNCMilterServer(self, port=port, address=bindaddress)
                     milterserver.serve()
                 else:
                     self.logger.debug("Starting MilterServer....")
-                    milterserver = MilterServer(
-                        self, port=port, address=bindaddress)
-
+                    milterserver = MilterServer(self, port=port, address=bindaddress)
                     tr = threading.Thread(target=milterserver.serve, args=())
                     tr.daemon = True
                     tr.start()
                 self.servers.append(milterserver)
             elif protocol == 'netcat':
-                ncserver = NCServer(
-                    self, port=port, address=bindaddress)
+                ncserver = NCServer(self, port=port, address=bindaddress)
                 tr = threading.Thread(target=ncserver.serve, args=())
                 tr.daemon = True
                 tr.start()
                 self.servers.append(ncserver)
             else:
-                self.logger.error(
-                    'Unknown Interface Protocol: %s, ignoring server on port %s' % (protocol, port))
+                self.logger.error('Unknown Interface Protocol: %s, ignoring server on port %s' % (protocol, port))
         except Exception as e:
-            self.logger.error(
-                "could not start connector %s/%s : %s" % (protocol, port, str(e)))
+            self.logger.error("could not start connector %s/%s : %s" % (protocol, port, str(e)))
 
     def _start_stats_thread(self):
         self.logger.info("Init Stat Engine")
@@ -690,22 +715,37 @@ class MainController(DefConfigMixin):
             minfreethreads = 0
 
         queuesize = maxthreads * 10
-        return ThreadPool(self,
-                          minthreads=minthreads, maxthreads=maxthreads,
-                          queuesize=queuesize, freeworkers=minfreethreads)
+        return ThreadPool(self, minthreads=minthreads, maxthreads=maxthreads, queuesize=queuesize, freeworkers=minfreethreads)
 
     def _start_processpool(self):
-        numprocs = self.config.getint('performance', 'initialprocs')
-        if numprocs < 1:
-            numprocs = multiprocessing.cpu_count() *2
+        numprocs = self.config.getfloat('performance', 'initialprocs')
+        if numprocs >= 1:
+            numprocs = int(numprocs)
+            self.logger.info(f"Set number of worker processors for procpool according to input to {numprocs}")
+        elif -1.0e-6 < numprocs < 1.0e-6:
+            self.logger.info(f"Set number of worker processors for procpool automatically (2*#procs) to {numprocs}")
+            numprocs = multiprocessing.cpu_count() * 2
+        else:
+            numprocs_in = numprocs
+            numprocs = max(int(multiprocessing.cpu_count() * abs(numprocs)), 1)
+            self.logger.info(f"Set number of worker processors for procpool automatically ({abs(numprocs_in)}*#procs) to {numprocs}")
+
         self.logger.info("Init process pool with %s worker processes"%(numprocs))
-        pool = fuglu.procpool.ProcManager(self._logQueue, numprocs = numprocs, config = self.config)
+        pool = fuglu.procpool.ProcManager(self._logQueue, numprocs=numprocs, config = self.config)
         return pool
 
     def _start_async_processpool(self):
-        numprocs = self.config.getint('performance', 'initialprocs')
-        if numprocs < 1:
+        numprocs = self.config.getfloat('performance', 'initialprocs')
+        if numprocs >= 1:
+            numprocs = int(numprocs)
+            self.logger.info(f"Set number of worker processors for async procpool according to input to {numprocs}")
+        elif -1.0e-6 < numprocs < 1.0e-6:
+            self.logger.info(f"Set number of worker processors for async procpool automatically (2*#procs) to {numprocs}")
             numprocs = multiprocessing.cpu_count() * 2
+        else:
+            numprocs_in = numprocs
+            numprocs = max(int(multiprocessing.cpu_count() * abs(numprocs)), 1)
+            self.logger.info(f"Set number of worker processors for async procpool automatically ({abs(numprocs_in)}*#procs) to {numprocs}")
 
         try:
             async_coroutines = self.config.getint('performance', 'async_coroutines', 10)
@@ -719,15 +759,13 @@ class MainController(DefConfigMixin):
 
     def _start_connectors(self):
         self.logger.info("Starting interface sockets...")
-        ports = self.config.get('main', 'incomingport')
-        for port in ports.split(','):
+        ports = self.config.getlist('main', 'incomingport')
+        for port in ports:
             self.start_connector(port)
 
     def _start_control_server(self):
-        control = ControlServer(self, address=self.config.get(
-            'main', 'bindaddress'), port=self.config.get('main', 'controlport'))
-        ctrl_server_thread = threading.Thread(
-            name='Control server', target=control.serve, args=())
+        control = ControlServer(self, address=self.config.get('main', 'bindaddress'), port=self.config.get('main', 'controlport'))
+        ctrl_server_thread = threading.Thread(name='Control server', target=control.serve, args=())
         ctrl_server_thread.daemon = True
         ctrl_server_thread.start()
         return control
@@ -775,8 +813,7 @@ class MainController(DefConfigMixin):
         self.load_extensions()
         ok = self.load_plugins()
         if not ok:
-            sys.stderr.write(
-                "Some plugins failed to load, please check the logs. Aborting.\n")
+            sys.stderr.write("Some plugins failed to load, please check the logs. Aborting.\n")
             self.logger.info('Fuglu shut down after fatal error condition')
             return EXIT_FATAL
 
@@ -793,6 +830,7 @@ class MainController(DefConfigMixin):
         self.controlserver = self._start_control_server()
 
         self.logger.info('Startup complete')
+        logging.getLogger('fuglu.main').info('Startup complete')
         self._run_main_loop()
         self.logger.info('Shutdown...')
         self.shutdown()
@@ -840,6 +878,7 @@ class MainController(DefConfigMixin):
 
             def readline(self):
                 return self.read(256)
+        
         sw = sw(clientsocket)
         sys.stdin = sw
         sys.stdout = sw
@@ -914,8 +953,7 @@ class MainController(DefConfigMixin):
             self.logger.error('backend not detected -> ignoring input!')
 
         # smtp engine changes?
-        ports = self.config.get('main', 'incomingport')
-        portspeclist = ports.split(',')
+        portspeclist = self.config.getlist('main', 'incomingport')
         portlist = []
 
         for portspec in portspeclist:
@@ -1241,7 +1279,7 @@ class MainController(DefConfigMixin):
         config for printing defaults/changed values
         """
         allOK = True
-        plugindirs = [dir for dir in self.config.get('main', 'plugindir').strip().split(',') if dir]
+        plugindirs = self.config.getlist('main', 'plugindir')
         for plugindir in plugindirs:
             if os.path.isdir(plugindir):
                 self.logger.debug('Searching for additional plugins in %s' % plugindir)
@@ -1252,23 +1290,20 @@ class MainController(DefConfigMixin):
 
         self.logger.debug('Module search path %s' % sys.path)
         self.logger.debug('Loading scanner plugins')
-        newplugins, loadok = self._load_all(self.config.get('main', 'plugins'))
+        newplugins, loadok = self._load_all(self.config.getlist('main', 'plugins'))
         if not loadok:
             allOK = False
 
-        newprependers, loadok = self._load_all(
-            self.config.get('main', 'prependers'))
+        newprependers, loadok = self._load_all(self.config.getlist('main', 'prependers'))
         if not loadok:
             allOK = False
 
-        newappenders, loadok = self._load_all(
-            self.config.get('main', 'appenders'))
+        newappenders, loadok = self._load_all(self.config.getlist('main', 'appenders'))
         if not loadok:
             allOK = False
 
         try:
-            milterplugins, loadok = self._load_all(
-                self.config.get('main', 'mplugins'))
+            milterplugins, loadok = self._load_all(self.config.getlist('main', 'mplugins'))
             if not loadok:
                 allOK = False
         except (configparser.NoSectionError, configparser.NoOptionError):
@@ -1295,13 +1330,11 @@ class MainController(DefConfigMixin):
 
         return allOK
 
-    def _load_all(self, configstring):
+    def _load_all(self, plugins):
         """load all plugins from config string. returns tuple ([list of loaded instances],allOk)"""
         pluglist = []
-        config_re = re.compile(
-            """^(?P<structured_name>[a-zA-Z0-9\.\_\-]+)(?:\((?P<config_override>[a-zA-Z0-9\.\_\-]+)\))?$""")
+        config_re = re.compile("""^(?P<structured_name>[a-zA-Z0-9._-]+)(?:\((?P<config_override>[a-zA-Z0-9._-]+)\))?$""")
         allOK = True
-        plugins = configstring.split(',')
         for plug in plugins:
             if plug == "":
                 continue
@@ -1313,16 +1346,17 @@ class MainController(DefConfigMixin):
             structured_name, configoverride = m.groups()
             structured_name = self.get_component_by_alias(structured_name)
             try:
-                plugininstance = self._load_component(
-                    structured_name, configsection=configoverride)
+                plugininstance = self._load_component(structured_name, configsection=configoverride)
                 pluglist.append(plugininstance)
             except (configparser.NoSectionError, configparser.NoOptionError):
                 CrashStore.store_exception()
                 self.logger.error("The plugin %s is accessing the config in __init__ -> can not load default values" % structured_name)
+            except (ValueError, ModuleNotFoundError) as e:
+                self.logger.error('Could not load plugin %s : %s' % (structured_name, e))
+                allOK = False
             except Exception as e:
                 CrashStore.store_exception()
-                self.logger.error('Could not load plugin %s : %s' %
-                                     (structured_name, e))
+                self.logger.error('Could not load plugin %s : %s' % (structured_name, e))
                 exc = traceback.format_exc()
                 self.logger.error(exc)
                 allOK = False
@@ -1341,11 +1375,10 @@ class MainController(DefConfigMixin):
             plugininstance = mod(self.config)
         else:
             # check if plugin supports config override
-            if 'section' in inspect.getargspec(mod.__init__)[0]:
+            if 'section' in inspect.getfullargspec(mod.__init__)[0]:
                 plugininstance = mod(self.config, section=configsection)
             else:
-                raise Exception('Cannot set Config Section %s : Plugin %s does not support config override' % (
-                    configsection, mod))
+                raise Exception('Cannot set Config Section %s : Plugin %s does not support config override' % (configsection, mod))
         return plugininstance
 
     def _propagate_plugin_defaults_to_config(self):
