@@ -15,24 +15,31 @@
 #
 # Fuglu SQLAlchemy Extension
 #
-from configparser import RawConfigParser
 from io import StringIO
 import logging
 import traceback
 import weakref
-from fuglu.shared import default_template_values
+from fuglu.shared import default_template_values, FuConfigParser, Suspect
+
 
 modlogger = logging.getLogger('fuglu.extensions.sql')
 STATUS = "not loaded"
 try:
-    from sqlalchemy import create_engine
+    from sqlalchemy import create_engine, text
     from sqlalchemy.orm import scoped_session, sessionmaker
-
+    from sqlalchemy.ext.declarative import declarative_base
+    DeclarativeBase = declarative_base()
+    
     SQL_EXTENSION_ENABLED = True
     STATUS = "available"
 except ImportError:
     SQL_EXTENSION_ENABLED = False
     STATUS = "sqlalchemy not installed"
+    DeclarativeBase = object
+
+    def text(string):
+        return string
+
 ENABLED = SQL_EXTENSION_ENABLED # fuglu compatibility
 
 
@@ -80,8 +87,7 @@ class DBFile(object):
             templatevars = {}
         sess = get_session(self.connectstring)
         res = sess.execute(self.query, templatevars)
-        self.logger.debug('Executing query %s with vars %s' %
-                          (self.query, templatevars))
+        self.logger.debug('Executing query %s with vars %s' % (self.query, templatevars))
         buff = []
         for row in res:
             line = " ".join(filter(None, row))
@@ -90,25 +96,39 @@ class DBFile(object):
         return buff
 
 
-class DBConfig(RawConfigParser):
+class DBConfig(FuConfigParser):
 
-    """Runtime Database Config Overrides. Behaves like a RawConfigParser but returns global database overrides/domainoverrides/useroverrides if available       
+    """
+    Runtime Database Config Overrides.
+    Behaves like a RawConfigParser but returns global database overrides/domainoverrides/useroverrides if available
     """
 
     def __init__(self, config, suspect):
-        RawConfigParser.__init__(self)
-
+        super().__init__()
+        self.logger = logging.getLogger('fuglu.sql.dbconfig')
+        self.sectioncache = {}
+        self.suspect = None
+        self.set_suspect(suspect)
+        self._clone_from(config)
+    
+    
+    def set_suspect(self, suspect):
+        if suspect is None:
+            suspect = Suspect('', '', '/dev/null')
         # store weak reference to suspect
         # otherwise (python 3), the instance of DBConfig does not reduce the
         # refcount to suspect even if it goes out of scope and the suspect
         # object does not get freed until a (manual or automatic) run of the
         # garbage collector "gc.collect()"
         self.suspect = weakref.ref(suspect)
-
-        self.logger = logging.getLogger('fuglu.sql.dbconfig')
-        self.cloneFrom(config)
-
-    def cloneFrom(self, config):
+    
+    
+    def set_rcpt(self, recipient):
+        suspect = Suspect('', recipient, '/dev/null')
+        self.set_suspect(suspect)
+    
+    
+    def _clone_from(self, config):
         """Clone this object from a RawConfigParser"""
         stringout = StringIO()
         config.write(stringout)
@@ -116,36 +136,86 @@ class DBConfig(RawConfigParser):
         del stringout
         self.read_file(stringin)
         del stringin
-
+    
+    
+    def load_section(self, section):
+        """
+        load section into local cache.
+        :param section: name of section to be loaded
+        :return: True if loading was successful
+        """
+        if not SQL_EXTENSION_ENABLED or (not self.has_section('databaseconfig')) or (not self.has_option('databaseconfig', 'dbconnectstring')):
+            return False
+        connectstring = self.parentget('databaseconfig', 'dbconnectstring')
+        if connectstring.strip() == '':
+            return False
+        query = self.parentget('databaseconfig', 'sqlsection')
+        if query.strip() == '':
+            return False
+        
+        session = get_session(connectstring)
+        sqlvalues = {'section': section}
+        default_template_values(self.suspect(), sqlvalues)
+        try:
+            result = session.execute(query, sqlvalues).fetchall()
+            self.sectioncache[section] = result
+        except Exception:
+            trb = traceback.format_exc()
+            self.logger.error("Error getting database config override section data: %s" % trb)
+            return False
+        return True
+    
+    
+    def _get_cached(self, section, option):
+        suspect = self.suspect()
+        todomain = f'%{suspect.to_domain}'
+        
+        optionfield = self.parentget('databaseconfig', 'option_field')
+        scopefield = self.parentget('databaseconfig', 'scope_field')
+        valuefield = self.parentget('databaseconfig', 'value_field')
+        
+        try:
+            for row in self.sectioncache.get(section, []):
+                scope = getattr(row, scopefield)
+                dboption = getattr(row, optionfield)
+                if dboption == option and scope in {suspect.to_address, todomain}:
+                    return getattr(row, valuefield)
+        except AttributeError as e:
+            self.logger.error('database layout does not match databaseconfig settings: %s' % str(e))
+        return None
+    
+    
     def get(self, section, option, **kwargs):
         if not SQL_EXTENSION_ENABLED or (not self.has_section('databaseconfig')) or (not self.has_option('databaseconfig', 'dbconnectstring')):
             return self.parentget(section, option, **kwargs)
-
+        
         connectstring = self.parentget('databaseconfig', 'dbconnectstring')
         if connectstring.strip() == '':
             #self.logger.debug('empty db connect string')
             return self.parentget(section, option, **kwargs)
-
-        session = get_session(connectstring)
+        
         query = self.parentget('databaseconfig', 'sql')
         if query.strip() == '':
             return self.parentget(section, option, **kwargs)
-
+        
+        if section in self.sectioncache:
+            return self._get_cached(section, option)
+        
+        session = get_session(connectstring)
         sqlvalues = {
             'section': section,
             'option': option,
         }
         default_template_values(self.suspect(), sqlvalues)
-
+        
         result = None
         try:
             #self.logger.debug("Executing query '%s' with vars %s"%(query,sqlvalues))
             result = session.execute(query, sqlvalues).first()
         except Exception:
             trb = traceback.format_exc()
-            self.logger.error(
-                "Error getting database config override: %s" % trb)
-
+            self.logger.error("Error getting database config override: %s" % trb)
+        
         session.remove()
         if result is None:
             #self.logger.debug('no result')
@@ -153,9 +223,10 @@ class DBConfig(RawConfigParser):
         else:
             #self.logger.debug('result: '+result[0])
             return result[0]
-
+    
+    
     def parentget(self, section, option, **kwargs):
-        return RawConfigParser.get(self, section, option, **kwargs)
+        return super().get(section, option, **kwargs)
 
 
 def get_domain_setting(domain, dbconnection, sqlquery, cache, cachename, default_value=None, logger=None):
@@ -177,12 +248,11 @@ def get_domain_setting(domain, dbconnection, sqlquery, cache, cachename, default
         dom = session.execute(sqlquery, {'domain': domain}).fetchall()
 
         if not dom or not dom[0] or len(dom[0]) == 0:
-            logger.debug(
-                "Can not load domain setting - domain %s not found. Using default settings." % domain)
+            logger.debug("Can not load domain setting - domain %s not found. Using default settings." % domain)
         else:
             settings = dom[0][0]
 
-        session.close()
+        session.remove()
 
     except Exception as e:
         logger.error("Exception while loading setting for %s : %s" % (domain, str(e)))
