@@ -13,6 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-__all__ = ['archive', 'attachment', 'actionoverride', 'clamav', 'sa', 'p_debug',
-           'p_skipper', 'p_fraction', 'vacation', 'icap', 'sssp', 'fprot', 'drweb', 'domainauth', 'script', 'a_statsd',
-           'fuzor', 'rspamd', 'killer', 'imapcopy', 'uriextract']
+__all__ = ['archive', 'attachment', 'decision', 'clamav', 'sa', 'p_debug', 'p_blwl', 'a_logging', 'bacn',
+           'p_skipper', 'p_fraction', 'vacation', 'antivirus', 'sssp', 'domainauth', 'script', 'a_statsd',
+           'fuzor', 'rspamd', 'mailcopy', 'uriextract', 'call_ahead', 'delay', 'ivmsg', 'domainauth',
+           'tlspolicy', 'originpolicy', 'outpolicy', 'restrictions', ]
