@@ -14,18 +14,34 @@
 # limitations under the License.
 #
 #
-from fuglu.shared import ScannerPlugin, DELETE, DUNNO, string_to_actioncode
+from fuglu.shared import ScannerPlugin, DELETE, DUNNO, string_to_actioncode, FileList, get_outgoing_helo
 from fuglu.bounce import Bounce
+from fuglu.stringencode import force_uString, force_bString
 from fuglu.extensions.sql import SQL_EXTENSION_ENABLED, DBFile, DBConfig
+from fuglu.extensions.redisext import RedisPooledConn, redis, ENABLED as REDIS_ENABLED
 from fuglu.extensions.filearchives import Archivehandle
 from fuglu.extensions.filetype import filetype_handler
-from fuglu.mailattach import NoExtractInfo
+from fuglu.mailattach import NoExtractInfo, Mailattachment
 import re
 import os
 import os.path
 import logging
 import email
+import hashlib
+import socket
+import json
 from threading import Lock
+try:
+    from domainmagic.rbl import RBLLookup
+    DOMAINMAGIC_AVAILABLE = True
+except ImportError:
+    DOMAINMAGIC_AVAILABLE=False
+try:
+    import kafka
+    KAFKA_AVAILABLE=True
+except ImportError:
+    KAFKA_AVAILABLE=False
+
 
 FUATT_NAMESCONFENDING = "-filenames.conf"
 FUATT_CTYPESCONFENDING = "-filetypes.conf"
@@ -33,29 +49,29 @@ FUATT_ARCHIVENAMESCONFENDING = "-archivenames.conf"
 FUATT_ARCHIVENAMES_CRYPTO_CONFENDING = "-archivenames-crypto.conf"
 FUATT_ARCHIVECTYPESCONFENDING = "-archivefiletypes.conf"
 
-FUATT_DEFAULT = u'default'
+FUATT_DEFAULT = 'default'
 
-FUATT_ACTION_ALLOW = u'allow'
-FUATT_ACTION_DENY = u'deny'
-FUATT_ACTION_DELETE = u'delete'
+FUATT_ACTION_ALLOW = 'allow'
+FUATT_ACTION_DENY = 'deny'
+FUATT_ACTION_DELETE = 'delete'
 
-FUATT_CHECKTYPE_FN = u'filename'
-FUATT_CHECKTYPE_CT = u'contenttype'
+FUATT_CHECKTYPE_FN = 'filename'
+FUATT_CHECKTYPE_CT = 'contenttype'
 
-FUATT_CHECKTYPE_ARCHIVE_CRYPTO_FN = u'archive-crypto-filename'
-FUATT_CHECKTYPE_ARCHIVE_FN = u'archive-filename'
-FUATT_CHECKTYPE_ARCHIVE_CT = u'archive-contenttype'
+FUATT_CHECKTYPE_ARCHIVE_CRYPTO_FN = 'archive-crypto-filename'
+FUATT_CHECKTYPE_ARCHIVE_FN = 'archive-filename'
+FUATT_CHECKTYPE_ARCHIVE_CT = 'archive-contenttype'
 
 ATTACHMENT_DUNNO = 0
 ATTACHMENT_BLOCK = 1
 ATTACHMENT_OK = 2
 ATTACHMENT_SILENTDELETE = 3
 
-KEY_NAME = u"name"
-KEY_CTYPE = u"ctype"
-KEY_ARCHIVENAME = u"archive-name"
-KEY_ARCHIVECTYPE = u"archive-ctype"
-KEY_ENCARCHIVENAME = u"enc-archive-name"  # name rules for files in password protected archives
+KEY_NAME = "name"
+KEY_CTYPE = "ctype"
+KEY_ARCHIVENAME = "archive-name"
+KEY_ARCHIVECTYPE = "archive-ctype"
+KEY_ENCARCHIVENAME = "enc-archive-name"  # name rules for files in password protected archives
 
 
 class RulesCache(object):
@@ -435,12 +451,11 @@ The other common template variables are available as well.
         self.checkarchivecontent = runtimeconfig.getboolean(self.section, 'checkarchivecontent')
         self.sendbounce = runtimeconfig.getboolean(self.section, 'sendbounce')
 
-        enabledarchivetypes = runtimeconfig.get(self.section, 'enabledarchivetypes')
+        enabledarchivetypes = runtimeconfig.getlist(self.section, 'enabledarchivetypes')
         if enabledarchivetypes:
-            enabled = [t.strip() for t in enabledarchivetypes.split(',')]
             archtypes = list(self.active_archive_extensions.keys())
             for archtype in archtypes:
-                if archtype not in enabled:
+                if archtype not in enabledarchivetypes:
                     del self.active_archive_extensions[archtype]
 
         returnaction = self.walk(suspect)
@@ -500,11 +515,11 @@ The other common template variables are available as well.
                 suspect.debug('%s Rulematch: Attachment=%s Rule=%s Description=%s Action=%s' % (
                     suspect.id, obj, regex, description, action))
                 if action == 'deny':
-                    self.logger.info('%s contains blocked attachment %s %s' % (
-                        suspect.id, displayname, asciirep))
-                    suspect.tags['blocked']['FiletypePlugin'] = True
+                    self.logger.info('%s contains blocked attachment %s %s' % (suspect.id, displayname, asciirep))
+                    blockinfo = {f'{displayname} {asciirep}': description}
+                    self._blockreport(suspect, blockinfo, enginename='FiletypePlugin')
                     blockinfo = ("%s %s: %s" % (displayname, asciirep, description)).strip()
-                    suspect.tags['FiletypePlugin.errormessage'] = blockinfo
+                    suspect.tags['FiletypePlugin.errormessage'] = blockinfo # deprecated
                     if self.sendbounce:
                         if suspect.is_spam() or suspect.is_virus():
                             self.logger.info(f"{suspect.id} backscatter prevention: not sending attachment block "
@@ -747,8 +762,8 @@ The other common template variables are available as well.
                                 # Keeping this check for backward compatibility
                                 # This could easily be removed since memory is used anyway
                                 if archivecontentmaxsize is not None and archObj.filesize > archivecontentmaxsize:
-                                    nocheckinfo.append(archObj.filename, u"toolarge",
-                                                       u"already extracted but too large for check: %u > %u"
+                                    nocheckinfo.append(archObj.filename, "toolarge",
+                                                       "already extracted but too large for check: %u > %u"
                                                        % (archObj.filesize, archivecontentmaxsize))
                                     continue
 
@@ -856,3 +871,568 @@ The other common template variables are available as well.
             print("No database configured. Using per user/domain file configuration from %s" %
                   self.config.get(self.section, 'rulesdir'))
         return True
+
+
+
+class AttHashAction(ScannerPlugin):
+    def __init__(self, config, section=None):
+        ScannerPlugin.__init__(self, config, section)
+        self.logger = self._logger()
+        self.rbllookup = None
+        self.requiredvars = {
+            'blacklistconfig':{
+                'default':'/etc/fuglu/atthash-rbl.conf',
+                'description':'Domainmagic RBL lookup config file',
+            },
+            'enginename': {
+                'default': '',
+                'description': 'set custom engine name',
+            },
+            'testhash': {
+                'default': 'E5NAEG57WZEJ4VGUOGEZ67NZ2FTD7RUV5QX6FIWEKOFKX5SR7UHQ',
+                'description': 'test record in database, used by lint',
+            },
+        }
+        
+    
+    def _init_rbllookup(self):
+        if self.rbllookup is None:
+            blacklistconfig = self.config.get(self.section,'blacklistconfig')
+            if os.path.exists(blacklistconfig):
+                self.rbllookup = RBLLookup()
+                self.rbllookup.from_config(blacklistconfig)
+                
+                
+    def _check_hash(self, myhash):
+        result = self.rbllookup.listings(myhash)
+        return result
+    
+    
+    def lint(self):
+        ok = self.check_config()
+        if ok and not DOMAINMAGIC_AVAILABLE:
+            print('ERROR: domainmagic not available - this plugin will do nothing')
+            ok = False
+        
+        if ok:
+            result = self._check_hash(self.config.get(self.section, 'testhash'))
+            if not 'Malware' in list(result.keys()):
+                print('ERROR: Eicar not detected as Malware')
+                ok = False
+        
+        return ok
+    
+    
+    def examine(self, suspect):
+        if not DOMAINMAGIC_AVAILABLE:
+            return DUNNO
+        self._init_rbllookup()
+        if self.rbllookup is None:
+            self.logger.error('Not scanning - blacklistconfig could not be loaded')
+            return DUNNO
+        
+        for attobj in suspect.att_mgr.get_objectlist(level=0):
+            filename = attobj.filename
+
+            if not attobj.is_attachment:
+                self.logger.debug('Skipping inline part with filename: %s' % filename)
+                continue
+
+            myhash = attobj.get_checksum('sha256', Mailattachment.HASHENC_B32, strip=True)
+            
+            result = self._check_hash(myhash)
+            for blockname in result.keys():
+                blockinfo = {filename: f'{blockname} {myhash}'}
+                enginename = self.config.get(self.section, 'enginename') or None
+                self._blockreport(suspect, blockinfo, enginename=enginename)
+                self.logger.info('%s attachment hash found: %s in %s' % (suspect.id, myhash, filename))
+                break
+            if not result:
+                self.logger.debug('%s no matching hash found for %s with hash %s' % (suspect.id, filename, myhash))
+            
+            if suspect.is_blocked():
+                break
+        else:
+            self.logger.debug('%s no attachment to check' % suspect.id)
+            
+        return DUNNO
+
+
+
+class FileHashCheck(ScannerPlugin):
+    
+    def __init__(self, config, section=None):
+        ScannerPlugin.__init__(self, config, section)
+        self.redis_pool = None
+        self.logger = logging.getLogger()
+        
+        self.requiredvars = {
+            'redis_conn': {
+                'default': '',
+                'description': 'redis backend database connection: redis://host:port/dbid',
+            },
+            'pinginterval': {
+                'default': '0',
+                'description': 'ping redis interval to prevent disconnect (0: don\'t ping)'
+            },
+            'timeout': {
+                'default': '2',
+                'description': 'redis/kafka timeout in seconds'
+            },
+            'hashtype': {
+                'default': 'MD5',
+                'description': 'the hashing algorithm to be used',
+            },
+            'extensionsfile': {
+                'default': '/etc/fuglu/conf.d/filehash_extensions.txt',
+                'description': 'path to file containing accepted file extensions. One per line, comments start after #',
+            },
+            'hashskiplistfile': {
+                'default': '/etc/fuglu/conf.d/filehash_skiphash.txt',
+                'description': 'path to file containing blacklisted hashes. One hash per line, comments start after #',
+            },
+            'filenameskipfile': {
+                'default': '/etc/fuglu/conf.d/filehash_skipfilename.txt',
+                'description': 'path to file containing blacklisted file name content. One per line, comments start after #',
+            },
+            'allowmissingextension': {
+                'default': 'False',
+                'description': 'check files without extensions',
+            },
+            'minfilesize': {
+                'default': '100',
+                'description': 'minimal size of a file to be checked',
+            },
+            'minfilesizebyext': {
+                'default': 'zip:40',
+                'description': 'comma separated list of file type specific min file size overrides. specifiy as ext:size',
+            }
+            
+        }
+        
+        self.minfilesizebyext = None
+        self.hashskiplist = None
+        self.extensions = None
+        self.filenameskip = None
+        self.allowmissingextension = False
+        self.minfilesize = 100
+    
+    
+    def _to_int(self, value, default=None):
+        try:
+            value = int(value)
+        except ValueError:
+            value = default
+        return value
+    
+    
+    def _init_databases(self):
+        if self.minfilesizebyext is None:
+            minfilesizebyext = self.config.getlist(self.section, 'minfilesizebyext')
+            self.minfilesizebyext = {}
+            for item in minfilesizebyext:
+                if not ':' in item:
+                    self.logger.error(f'minfilesizebyext {item} is not a valid specification')
+                k, v = item.split(':', 1)
+                try:
+                    v = int(v)
+                except ValueError:
+                    self.logger.error(f'minfilesizebyext value {v} for {k} is not an integer')
+                    continue
+                self.minfilesizebyext[k.lower()] = v
+        if self.extensions is None:
+            filepath = self.config.get(self.section, 'extensionsfile')
+            if filepath and os.path.exists(filepath):
+                self.extensions = FileList(filename=filepath,
+                    lowercase=True, additional_filters=[FileList.inline_comments_filter])
+            else:
+                self.logger.error(f'extensionfile {filepath} does not exist')
+        if self.hashskiplist is None:
+            filepath = self.config.get(self.section, 'hashskiplistfile')
+            if filepath and os.path.exists(filepath):
+                self.hashskiplist = FileList(filename=filepath,
+                    lowercase=True, additional_filters=[FileList.inline_comments_filter])
+            else:
+                self.logger.error(f'hashskiplistfile {filepath} does not exist')
+        if self.filenameskip is None:
+            filepath = self.config.get(self.section, 'filenameskipfile')
+            if filepath and os.path.exists(filepath):
+                self.filenameskip = FileList(filename=filepath,
+                    lowercase=True, additional_filters=[FileList.inline_comments_filter])
+            else:
+                self.logger.error(f'filenameskipfile {filepath} does not exist')
+        self.allowmissingextension = self.config.getboolean(self.section, 'allowmissingextension')
+        self.minfilesize = self.config.getint(self.section, 'minfilesize')
+    
+    
+    def examine(self, suspect):
+        if not REDIS_ENABLED:
+            return DUNNO
+        
+        hashtype = self.config.get(self.section, 'hashtype')
+        hashtype = hashtype.lower()
+        if not hasattr(hashlib, hashtype):
+            self.logger.error('invalid hash type %s' % hashtype)
+        
+        self._init_databases()
+        
+        if self.hashskiplist:
+            hashskiplist = self.hashskiplist.get_list()
+        else:
+            hashskiplist = []
+        
+        for attobj in suspect.att_mgr.get_objectlist(level=0):
+            filename = attobj.filename
+            filecontent = attobj.buffer if attobj.buffer else b""
+            
+            if not attobj.is_attachment:
+                self.logger.debug('Skipping inline part with filename: %s' % filename)
+                continue
+            
+            if not self._check_filename(filename, len(filecontent)):
+                self.logger.debug('Skipping attachment size %s with filename: %s' % (len(filecontent), filename))
+                continue
+            
+            myhash = attobj.get_checksum(hashtype)
+            if myhash in hashskiplist:
+                self.logger.debug('Skiplisted hash: %s' % myhash)
+                continue
+            
+            virusname = self._check_hash(myhash)
+            if virusname is not None:
+                blockinfo = {filename: f'{virusname} {myhash}'}
+                self._blockreport(suspect, blockinfo)
+                break
+            else:
+                self.logger.debug('no matching hash found for %s with hash %s' % (filename, myhash))
+        
+        return DUNNO, None
+    
+    
+    def _check_skip(self, suspect):
+        skip = None
+        if suspect.is_blocked():
+            skip = 'blocked'
+        elif suspect.is_spam():
+            skip = 'spam'
+        elif suspect.is_virus():
+            skip = 'virus'
+        return skip
+    
+    
+    def _lint_redis(self):
+        success = True
+        redis_conn = self.config.get(self.section, 'redis_conn')
+        if redis_conn and not REDIS_ENABLED:
+            print('ERROR: redis not available')
+            success = False
+        elif redis_conn:
+            self._init_redis()
+            if self.redis_pool is None:
+                success = False
+                print('ERROR: could not connect to redis server: %s' % redis_conn)
+            else:
+                try:
+                    reply = self.redis_pool.check_connection()
+                    if reply:
+                        print('OK: redis server replied to ping')
+                    else:
+                        print('ERROR: redis server did not reply to ping')
+                
+                except redis.exceptions.ConnectionError as e:
+                    success = False
+                    print('ERROR: failed to talk to redis server: %s' % str(e))
+        else:
+            print('INFO: redis disabled')
+        return success
+    
+    
+    def lint(self):
+        success = self.check_config()
+        if not success:
+            return success
+        
+        success = self._lint_redis()
+        
+        hashtype = self.config.get(self.section, 'hashtype')
+        hashtype = hashtype.lower()
+        if not hasattr(hashlib, hashtype):
+            print('ERROR: invalid hash type %s' % hashtype)
+            success = False
+        
+        self.hashskiplist = FileList(filename=self.config.get(self.section, 'hashblacklistfile'), lowercase=True,
+                                     additional_filters=[FileList.inline_comments_filter])
+        if self.hashskiplist.get_list():
+            print('WARNING: empty hash skiplist')
+        
+        self.extensions = FileList(filename=self.config.get(self.section, 'extensionsfile'), lowercase=True,
+                                   additional_filters=[FileList.inline_comments_filter])
+        ext = self.extensions.get_list()
+        if len(ext) == 0:
+            success = False
+            print('WARNING: extensions list is empty')
+        else:
+            print('INFO: checking %s extensions' % len(ext))
+        
+        self.filenameskip = FileList(filename=self.config.get(self.section, 'bannedfilenamecontentfile'),
+                                     lowercase=True, additional_filters=[FileList.inline_comments_filter])
+        if len(self.filenameskip.get_list()) == 0:
+            success = False
+            print('WARNING: extensions list is empty')
+        
+        return success
+    
+    
+    def _init_redis(self):
+        """
+        initializes Redis on first call and returns the Redis connection object. Aborts the program if redis configuration is malformed.
+        :return: (RedisKeepAlive)
+        """
+        if self.redis_pool is None:
+            redis_conn = self.config.get(self.section, 'redis_conn')
+            if redis_conn:
+                timeout = self.config.getint(self.section, 'timeout'),
+                pinginterval = self.config.getint(self.section, 'pinginterval')
+                self.redis_pool = RedisPooledConn(redis_conn, socket_keepalive=True, socket_timeout=timeout, pinginterval=pinginterval)
+    
+    
+    def _check_hash(self, myhash):
+        virusname = None
+        self._init_redis()
+        
+        if self.redis_pool is not None:
+            attempts = 2
+            while attempts:
+                attempts -= 1
+                try:
+                    self.redis_pool: RedisPooledConn
+                    redisconn = self.redis_pool.get_conn()
+                    result = redisconn.hmget(myhash, ['virusname'])
+                    if result:
+                        virusname = force_uString(result[0])
+                    attempts = 0
+                except redis.exceptions.ConnectionError as e:
+                    msg = f"problem in {self.__class__.__name__} getting virusname for haxh {myhash}"
+                    self.logger.warning(msg) if attempts else self.logger.error(msg, exc_info=e)
+                except (socket.timeout, redis.exceptions.TimeoutError):
+                    self.logger.info("Socket timeout in check_hash")
+        return virusname
+    
+    
+    def _check_filename(self, filename, filesize, force=False, ignore_upn=False, ignore_dotfiles=False):
+        ok = True
+        
+        if ignore_upn:
+            for i in filename:
+                if ord(i) > 128:
+                    self.logger.debug('skipping file %s - name contains upn' % filename)
+                    ok = False
+                    break
+        
+        if ok and filename in ['unnamed.htm', 'unnamed.txt']:  # ignore text and html parts of mail
+            ok = False
+        
+        if ok and ignore_dotfiles and filename.startswith('.'):
+            self.logger.debug('skipping file %s - is hidden' % filename)
+            ok = False
+        
+        lowerfile = filename.lower()
+        try:
+            ext = lowerfile.rsplit('.', 1)[1]
+        except IndexError:
+            ext = ''
+        
+        if ok and self.extensions and ext != '':
+            if ext not in self.extensions.get_list() and not force:
+                self.logger.debug('skipping file %s - extension not in my list' % filename)
+                ok = False
+        elif ok:
+            if not self.allowmissingextension and not force:
+                self.logger.debug('skipping file %s with missing extension' % filename)
+                ok = False
+        
+        if ok and self.filenameskip and not force:
+            for badword in self.filenameskip.get_list():
+                if badword in lowerfile:
+                    self.logger.debug("filename %s contains bad word '%s' - skipping" % (filename, badword))
+                    ok = False
+                    break
+            
+            if ok and ext in self.minfilesizebyext:
+                if filesize < self.minfilesizebyext[ext]:
+                    self.logger.debug('%s too small for extension %s (%s bytes)' % (filename, ext, filesize))
+                    ok = False
+            elif ok and filesize < self.minfilesize:
+                self.logger.debug('ignoring small file %s (%s bytes)' % (filename, filesize))
+                ok = False
+        
+        return ok
+
+
+class FileHashRedis(object):
+    def __init__(self, redis_pool, ttl, logger):
+        self.redis_pool = redis_pool
+        self.ttl = ttl
+        self.logger = logger
+    
+    def insert(self, myhash, messagebytes, age=0):
+        values = json.loads(messagebytes)
+        filename = values.get('filename')
+        virusname = values.get('virusname')
+        filesize = values.get('filesize')
+        
+        td = int(self.ttl-age)
+        if td <= 0:
+            self.logger.debug(f'skipping old hash {myhash} with age {age}')
+            return
+        
+        try:
+            redisconn = self.redis_pool.get_conn()
+            result = redisconn.hmget(myhash, ['virusname', 'filesize'])
+            if result and result[1] == filesize:
+                virusname = result[0]
+                self.logger.debug('known hash %s has virus name %s' % (myhash, virusname))
+            else:
+                redisconn.hmset(myhash, dict(filename=filename, filesize=filesize, virusname=virusname))
+            redisconn.expire(myhash, td)
+        except redis.exceptions.ConnectionError as e:
+            self.logger.exception(e)
+        except (socket.timeout, redis.exceptions.TimeoutError):
+            self.logger.info("Socket timeout in insert_redis")
+
+
+class FileHashFeeder(FileHashCheck):
+    def __init__(self,config,section=None):
+        FileHashCheck.__init__(self,config,section)
+        self.logger=logging.getLogger()
+        
+        requiredvars = {
+            'prefix': {
+                'default': 'FH.GEN',
+                'description': 'virus name prefix',
+            },
+            'expirationdays': {
+                'default':'3',
+                'description':'days until hash expires',
+            },
+            'kafkahosts': {
+                'default': '',
+                'description:': 'kafka bootstrap hosts: host1:port host2:port'
+            },
+            'kafkatopic': {
+                'default': 'filehash',
+                'description': 'name of kafka topic'
+            },
+            'kafkausername': {
+                'default': '',
+                'description:': 'kafka sasl user name for this producer'
+            },
+            'kafkapassword': {
+                'default': '',
+                'description': 'kafka sals password for this producer'
+            },
+        }
+        self.requiredvars.update(requiredvars)
+        
+        self.kafkaproducer = None
+        self.kafkatopic = None
+        self.delta_expiration = 0
+
+
+    def lint(self):
+        success = FileHashCheck.lint(self)
+        expirationdays = self.config.get(self.section, 'expirationdays')
+        try:
+            int(expirationdays)
+        except ValueError:
+            success = False
+            print('ERROR: expirationdays must be a number. current value: %s' % expirationdays)
+        if self.config.get(self.section, 'kafkahosts'):
+            try:
+                self._init_kafka()
+            except kafka.errors.KafkaError as e:
+                print('ERROR: failed to connect to kafka: %s' % str(e))
+                success = False
+            except Exception as e:
+                print('ERROR: Error connecting to kafka: %s' % str(e))
+                self.logger.exception(e)
+                success = False
+        else:
+            print('INFO: kafka disabled')
+        return success
+
+    def process(self, suspect, decision):
+        self._run(suspect)
+    
+    
+    def examine(self, suspect):
+        self._run(suspect)
+        return DUNNO, None
+    
+    
+    def _run(self, suspect):
+        hashtype = self.config.get(self.section, 'hashtype')
+        hashtype = hashtype.lower()
+        if not hasattr(hashlib, hashtype):
+            self.logger.error('invalid hash type %s' % hashtype)
+
+        self._init_databases()
+        self.delta_expiration = self.config.getint(self.section, 'expirationdays') * 86400
+        
+        if self.hashskiplist:
+            hashskiplist = self.hashskiplist.get_list()
+        else:
+            hashskiplist = []
+        
+        for attobj in suspect.att_mgr.get_objectlist(level=0):
+            filename = attobj.filename
+            filecontent = attobj.buffer if attobj.buffer else b""
+            if not self._check_filename(filename, len(filecontent)):
+                continue
+
+            myhash = attobj.get_checksum(hashtype)
+            if myhash in hashskiplist:
+                self.logger.debug('Skiplisted hash: %s' % myhash)
+                continue
+            
+            virusname = attobj.get_mangled_filename(prefix=self.config.get(self.section, 'prefix'))
+            messagebytes = json.dumps({'filename':filename, 'virusname':virusname, 'filesize':len(filecontent), 'filehash':myhash}).encode()
+            if self.config.get(self.section, 'redis_conn'):
+                self._insert_redis(myhash, messagebytes)
+            if self.config.get(self.section, 'kafkahosts'):
+                try:
+                    self._insert_kafka(myhash, messagebytes)
+                    self.logger.info('%s logged hash %s to kafka' % (suspect.id, myhash))
+                except kafka.errors.KafkaError as e:
+                    self.logger.error('%s failed to log hash %s due to %s' % (suspect.id, myhash, str(e)))
+    
+    
+    def _insert_redis(self, myhash, messagebytes):
+        self._init_redis()
+        redisbackend = FileHashRedis(self.redis_pool, self.delta_expiration, self.logger)
+        redisbackend.insert(myhash, messagebytes, 0)
+    
+    
+    def _init_kafka(self):
+        if self.kafkaproducer is not None:
+            return
+        self.bootstrap_servers = self.config.get(self.section, 'kafkahosts').split()
+        if self.bootstrap_servers:
+            self.kafkatopic = self.config.get(self.section, 'kafkatopic')
+            timeout = self.config.getint(self.section, 'timeout')
+            username = self.config.get(self.section, 'kafkausername')
+            password = self.config.get(self.section, 'kafkapassword')
+            clientid = 'prod-fuglu-%s-%s' % (self.__class__.__name__, get_outgoing_helo(self.config))
+            self.kafkaproducer = kafka.KafkaProducer(bootstrap_servers=self.bootstrap_servers, api_version=(0, 10, 1), client_id=clientid,
+                                                     request_timeout_ms=timeout*1000, sasl_plain_username=username, sasl_plain_password=password)
+    
+    
+    def _insert_kafka(self, myhash, messagebytes):
+        if self.kafkaproducer is None:
+            self._init_kafka()
+        try:
+            self.kafkaproducer.send(self.kafkatopic, value=messagebytes, key=force_bString(myhash))
+        except Exception as e:
+            self.kafkaproducer = None
+            raise e
+
