@@ -21,18 +21,25 @@ import logging
 import socket
 import os
 from email import message_from_bytes
-from fuglu.shared import ScannerPlugin, AppenderPlugin, SuspectFilter, DUNNO, FileList, HOSTNAME
-from fuglu.extensions.redisext import RedisKeepAlive, redis, ENABLED as REDIS_ENABLED
+from fuglu.shared import ScannerPlugin, AppenderPlugin, SuspectFilter, DUNNO, FileList, get_outgoing_helo
+from fuglu.extensions.redisext import RedisPooledConn, redis, ENABLED as REDIS_ENABLED
 from fuglu.lib.patchedemail import PatchedMessage
 from fuglu.stringencode import force_bString
 
+CONN_EXC = []
+if REDIS_ENABLED:
+    CONN_EXC.append(redis.exceptions.ConnectionError)
+
 try:
     # requires kafka-python
     import kafka
+    CONN_EXC.append(kafka.errors.KafkaError)
     KAFKA_AVAILABLE=True
 except ImportError:
     KAFKA_AVAILABLE=False
-    
+CONN_EXC = tuple(CONN_EXC)
+
+
 
 
 class FuzorMixin(object):
@@ -43,15 +50,15 @@ class FuzorMixin(object):
                 'description': 'storage backend to be used. currently supported: redis (check/report), kafka (report only)'
             },
             
-            'redis': {
-                'default': 'localhost:6379:0',
-                'description': 'redis config: host:port:db',
+            'redis_conn': {
+                'default': 'redis://localhost:6379/0',
+                'description': 'redis config: redis://host:port/db',
             },
-            'redispw': {
+            'redis_password': {
                 'default': '',
-                'description': 'password to connect to redis database. leave empty for no password',
+                'description': 'password to connect to redis database. leave empty for no password. deprecated, set password in redis_conn url',
             },
-            'ttl': {
+            'redis_ttl': {
                 'default': '604800',
                 'description': 'hash redis ttl in seconds',
             },
@@ -70,7 +77,7 @@ class FuzorMixin(object):
             },
             'kafkapassword': {
                 'default': '',
-                'description': 'kafka sals password for this producer'
+                'description': 'kafka sasl password for this producer'
             },
             
             'maxsize': {
@@ -93,29 +100,24 @@ class FuzorMixin(object):
         
     def _init_backend(self):
         backend = self.config.get(self.section, 'backend').lower()
+        
         try:
             if backend == 'redis':
                 self._init_backend_redis()
             elif backend == 'kafka':
                 self._init_backend_kafka()
-        except (kafka.errors.KafkaError, redis.exceptions.ConnectionError) as e:
+        except CONN_EXC as e:
             raise BackendError('backend=%s, error=%s' % (backend, str(e)))
         
         
     def _init_backend_redis(self):
         if self.backend is not None:
             return
-        host, port, db = self.config.get(self.section, 'redis').split(':')
-        password = self.config.get(self.section, 'redispw') or None
-        red = RedisKeepAlive(
-            host=host,
-            port=port,
-            db=int(db),
-            password=password,
-            socket_keepalive=True,
-            socket_timeout=self.config.getint(self.section, 'timeout'))
-        self.backend = RedisBackend(red)
-        self.backend.ttl = self.config.getint(self.section, 'ttl')
+        redis_url = self.config.get(self.section, 'redis_conn')
+        password = self.config.get(self.section, 'redis_password') or None
+        redis_pool = RedisPooledConn(redis_url, password=password)
+        self.backend = RedisBackend(redis_pool)
+        self.backend.ttl = self.config.getint(self.section, 'redis_ttl')
     
     
     def _init_backend_kafka(self):
@@ -179,8 +181,7 @@ class FuzorMixin(object):
         if ok:
             try:
                 self._init_backend()
-        
-                reply = self.backend.redis.ping()
+                reply = self.backend.redis_pool.check_connection()
                 if reply:
                     print('OK: redis server replied to ping')
                 else:
@@ -247,16 +248,26 @@ class FuzorMixin(object):
             except BackendError as e:
                 self.logger.error('failed to connect to backend: %s' % str(e))
                 have_backend = False
-                
-            try:
-                if have_backend:
+
+            attempts = 2
+            while attempts:
+                attempts -= 1
+                try:
+                    self._init_backend()
                     count = self.backend.increase(fuhash.digest)
                     self.logger.info("suspect %s hash %s seen %s times before" % (suspect.id, fuhash.digest, count - 1))
-            except redis.exceptions.TimeoutError as e:
-                self.logger.error('%s failed increasing count due to %s' % (suspect.id, str(e)))
-            except (redis.exceptions.ConnectionError, kafka.errors.KafkaError) as e:
-                self.logger.error('%s failed increasing count due to %s, resetting connection' % (suspect.id, str(e)))
-                self.backend = None
+                    attempts = 0
+                except redis.exceptions.TimeoutError as e:
+                    msg = f'{suspect.id} failed increasing count due to {str(e)}'
+                    self.logger.warning(msg) if attempts else self.logger.error(msg)
+                except BackendError as e:
+                    msg = f'{suspect.id} failed to connect to backend: {str(e)}, resetting connection'
+                    self.backend = None
+                    self.logger.warning(msg) if attempts else self.logger.error(msg)
+                except CONN_EXC as e:
+                    msg = f'{suspect.id} failed increasing count due to {str(e)}, resetting connection'
+                    self.backend = None
+                    self.logger.warning(msg) if attempts else self.logger.error(msg)
         else:
             self.logger.info("suspect %s not enough data for a digest" % suspect.id)
         
@@ -284,6 +295,8 @@ class FuzorReport(ScannerPlugin, FuzorMixin):
         digest, count = self.report(suspect)
         if digest is not None and suspect.get_tag('FuZor') is None:
             suspect.set_tag('FuZor', (digest, count))
+            suspect.set_tag('fuzor.digest', digest)
+            suspect.set_tag('fuzor.count', count)
         return DUNNO
     
     
@@ -295,8 +308,6 @@ class FuzorReport(ScannerPlugin, FuzorMixin):
 class FuzorReportAppender(AppenderPlugin, FuzorMixin):
     """ Report all messages to the fuzor redis backend"""
     
-    
-    
     def __init__(self, config, section=None):
         AppenderPlugin.__init__(self, config, section)
         FuzorMixin.__init__(self)
@@ -311,6 +322,8 @@ class FuzorReportAppender(AppenderPlugin, FuzorMixin):
         digest, count = self.report(suspect)
         if digest is not None and suspect.get_tag('FuZor') is None:
             suspect.set_tag('FuZor', (digest, count))
+            suspect.set_tag('fuzor.digest', digest)
+            suspect.set_tag('fuzor.count', count)
     
     
     def lint(self):
@@ -375,34 +388,50 @@ class FuzorCheck(ScannerPlugin, FuzorMixin):
             msg = suspect.get_message_rep()
         # self.logger.info("%s: FUZOR PRE-HASH"%suspect.id)
         fuhash = FuzorDigest(msg)
+        digest = fuhash.digest
         # self.logger.info("%s: FUZOR POST-HASH"%suspect.id)
-        if fuhash.digest is not None:
-            suspect.debug('Fuzor digest = %s' % fuhash.digest)
+        if digest is not None:
+            suspect.debug('Fuzor digest = %s' % digest)
             
             self._init_ignorelist()
             if self.ignorelist is not None:
                 ignorelist = self.ignorelist.get_list()
-                if fuhash.digest in ignorelist:
-                    self.logger.info('%s ignoring fuzor hash %s' % (suspect.id, fuhash.digest))
+                if digest in ignorelist:
+                    self.logger.info('%s ignoring fuzor hash %s' % (suspect.id, digest))
                     return DUNNO
             
-            # self.logger.info("%s: FUZOR INIT-BACKEND"%suspect.id)
-            try:
-                self._init_backend()
-            except BackendError as e:
-                self.logger.error('%s failed to initialise backend: %s' % (suspect.id, str(e)))
-                return DUNNO
-            
-            # self.logger.info("%s: FUZOR START-QUERY"%suspect.id)
-            try:
-                count = self.backend.get(fuhash.digest)
-            except redis.exceptions.TimeoutError as e:
-                self.logger.error('%s failed getting count due to %s' % (suspect.id, str(e)))
-                return DUNNO
-            except redis.exceptions.ConnectionError as e:
-                self.logger.error('%s failed getting count due to %s, resetting connection' % (suspect.id, str(e)))
-                self.backend = None
-                return DUNNO
+            count = 0
+            attempts = 2
+            while attempts:
+                attempts -= 1
+                # self.logger.info("%s: FUZOR INIT-BACKEND"%suspect.id)
+                # self.logger.info("%s: FUZOR START-QUERY"%suspect.id)
+                try:
+                    self._init_backend()
+                    count = self.backend.get(digest)
+                except redis.exceptions.TimeoutError as e:
+                    msg = f"{suspect.id} failed getting count due to {str(e)}"
+                    if attempts:
+                        self.logger.warning(msg)
+                    else:
+                        self.logger.error(msg)
+                        return DUNNO
+                except BackendError as e:
+                    msg = f'{suspect.id} failed to initialise backend: {str(e)}, resetting connection'
+                    self.backend = None
+                    if attempts:
+                        self.logger.warning(msg)
+                    else:
+                        self.logger.error(msg)
+                        return DUNNO
+                except redis.exceptions.ConnectionError as e:
+                    msg = f'{suspect.id} failed getting count due to {str(e)}, resetting connection'
+                    self.backend = None
+                    if attempts:
+                        self.logger.warning(msg)
+                    else:
+                        self.logger.error(msg)
+                        return DUNNO
 
             # self.logger.info("%s: FUZOR END-QUERY"%suspect.id)
             headername = self.config.get(self.section, 'headername')
@@ -411,14 +440,16 @@ class FuzorCheck(ScannerPlugin, FuzorMixin):
             #     self._writeheader(suspect,headername,'HIGH')
             # elif count>self.config.getint(self.section,'lowthreshold'):
             #     self._writeheader(suspect,headername,'LOW')
-            suspect.set_tag('FuZor', (fuhash.digest, count))
+            suspect.set_tag('FuZor', (digest, count))
+            suspect.set_tag('fuzor.digest', digest)
+            suspect.set_tag('fuzor.count', count)
             if count > 0:
                 # self.logger.info("%s: FUZOR WRITE HEADER"%suspect.id)
-                suspect.write_sa_temp_header("%s-ID" % headername, fuhash.digest)
+                suspect.write_sa_temp_header("%s-ID" % headername, digest)
                 suspect.write_sa_temp_header("%s-Lvl" % headername, count)
-                self.logger.info("%s digest %s seen %s times" % (suspect.id, fuhash.digest, count))
+                self.logger.info("%s digest %s seen %s times" % (suspect.id, digest, count))
             else:
-                self.logger.debug("%s digest %s not seen before" % (suspect.id, fuhash.digest))
+                self.logger.debug("%s digest %s not seen before" % (suspect.id, digest))
         else:
             suspect.debug('%s not enough data for a digest' % suspect.id)
         
@@ -595,13 +626,14 @@ class FuzorBackend(object):
 
 
 class RedisBackend(FuzorBackend):
-    def __init__(self, redisconn=None):
-        self.redis = redisconn or redis.StrictRedis()
+    def __init__(self, redispool=None):
+        self.redis_pool = redispool or RedisPooledConn()
         self.ttl = 7 * 24 * 3600
         self.logger = logging.getLogger("fuglu.fuzor.RedisBackend")
     
     def increase(self, digest):
-        pipe = self.redis.pipeline()
+        redisconn = self.redis_pool.get_conn()
+        pipe = redisconn.pipeline()
         pipe.incr(digest)
         pipe.expire(digest, self.ttl)
         try:
@@ -613,7 +645,8 @@ class RedisBackend(FuzorBackend):
     
     def get(self, digest):
         try:
-            return int(self.redis.get(digest))
+            redisconn = self.redis_pool.get_conn()
+            return int(redisconn.get(digest))
         except (ValueError, TypeError) as e:
             return 0
         except (socket.timeout, redis.exceptions.TimeoutError):
@@ -625,7 +658,7 @@ class RedisBackend(FuzorBackend):
 class KafkaBackend(FuzorBackend):
     def __init__(self, bootstrap_servers, topic, timeout, username, password):
         self.logger = logging.getLogger("fuglu.fuzor.KafkaBackend")
-        clientid = 'prod-fuglu-%s-%s' % (self.__class__.__name__, HOSTNAME)
+        clientid = 'prod-fuglu-%s-%s' % (self.__class__.__name__, get_outgoing_helo())
         self.producer = kafka.KafkaProducer(
             bootstrap_servers=bootstrap_servers, api_version=(0, 10, 1), client_id=clientid,
             request_timeout_ms=timeout*1000, sasl_plain_username=username, sasl_plain_password=password
