@@ -22,11 +22,11 @@ import os
 class PluginSkipper(PrependerPlugin):
 
     """Skips plugins based on standard filter file
-This can be used for example to skip spam filters on outgoing messages.
-e.g. put this in /etc/fuglu/skipplugins.regex:
-
-@incomingport    1099    SAPlugin
-"""
+    This can be used for example to skip spam filters on outgoing messages.
+    e.g. put this in /etc/fuglu/skipplugins.regex:
+    
+    @incomingport    1099    SAPlugin
+    """
 
     def __init__(self, config, section=None):
         PrependerPlugin.__init__(self, config, section)
@@ -34,33 +34,39 @@ e.g. put this in /etc/fuglu/skipplugins.regex:
         self.requiredvars = {
             'filterfile': {
                 'default': '/etc/fuglu/skipplugins.regex',
+                'description': 'path to file containing scanner plugin skip regex rules'
             }
         }
         self.logger = self._logger()
-
+    
+    
     def __str__(self):
         return "Plugin Skipper"
-
+    
+    
     def pluginlist(self, suspect, pluginlist):
         """Removes scannerplugins based on filter file"""
         if not self._initfilter():
             return None
 
-        args = self.filter.get_args(suspect)
+        argsext = self.filter.get_args(suspect, True)
         # each arg should be a comma separated list of classnames to skip
 
-        skippluginlist = []
-
-        for arg in args:
-            skippluginlist.extend(arg.split(','))
+        skipplugindict = {}
+        for fieldname, strval, arg, pattern in argsext:
+            plugnames = arg.split(',')
+            for plugname in plugnames:
+                skipplugindict[plugname] = fieldname
 
         listcopy = pluginlist[:]
         for plug in pluginlist:
             name = plug.__class__.__name__
-            if name in skippluginlist:
+            if name in skipplugindict:
                 listcopy.remove(plug)
+                suspect.set_tag(f'{name}.skipreason', skipplugindict[name])
         return listcopy
-
+    
+    
     def _initfilter(self):
         if self.filter is not None:
             return True
@@ -75,14 +81,16 @@ e.g. put this in /etc/fuglu/skipplugins.regex:
 
         self.filter = SuspectFilter(filename)
         return True
-
+    
+    
     def lint(self):
         return self.check_config() and self.lint_filter()
-
+    
+    
     def lint_filter(self):
         filterfile = self.config.get(self.section, 'filterfile')
-        filter = SuspectFilter(filterfile)
-        return filter.lint()
+        sfilter = SuspectFilter(filterfile)
+        return sfilter.lint()
 
 
 class AppenderSkipper(PrependerPlugin):
@@ -102,13 +110,16 @@ class AppenderSkipper(PrependerPlugin):
         self.requiredvars = {
             'filterfile': {
                 'default': '/etc/fuglu/skipappenders.regex',
+                'description': 'path to file containing appender plugin skip regex rules'
             }
         }
         self.logger = self._logger()
-
+    
+    
     def __str__(self):
         return "Appender Plugin Skipper"
-
+    
+    
     def appenderlist(self, suspect, pluginlist):
         """Removes appender plugins based on filter file"""
         if not self._initfilter():
@@ -128,7 +139,8 @@ class AppenderSkipper(PrependerPlugin):
             if name in skippluginlist:
                 listcopy.remove(plug)
         return listcopy
-
+    
+    
     def _initfilter(self):
         if self.filter is not None:
             return True
@@ -143,11 +155,13 @@ class AppenderSkipper(PrependerPlugin):
 
         self.filter = SuspectFilter(filename)
         return True
-
+    
+    
     def lint(self):
         return self.check_config() and self.lint_filter()
-
+    
+    
     def lint_filter(self):
         filterfile = self.config.get(self.section, 'filterfile')
-        filter = SuspectFilter(filterfile)
-        return filter.lint()
+        sfilter = SuspectFilter(filterfile)
+        return sfilter.lint()
