@@ -14,16 +14,15 @@
 # limitations under the License.
 #
 #
-from fuglu.shared import ScannerPlugin, DUNNO, DEFER, Suspect, string_to_actioncode, apply_template
-from fuglu.extensions.sql import DBConfig, get_session, SQL_EXTENSION_ENABLED
+from fuglu.shared import ScannerPlugin, DUNNO, Suspect, string_to_actioncode, apply_template
+from fuglu.extensions.sql import DBConfig, get_session, SQL_EXTENSION_ENABLED, DeclarativeBase
 from fuglu.stringencode import force_bString, force_uString
-from string import Template
+from fuglu.lib.patchedemail import PatchedMessage
 import time
 import socket
 import email
 import re
 import os
-from fuglu.lib.patchedemail import PatchedMessage
 from email.mime.text import MIMEText
 
 GTUBE = """Date: Mon, 08 Sep 2008 17:33:54 +0200
@@ -34,6 +33,20 @@ Subject: test scanner
   XJS*C4JDBQADN1.NSBN3*2IDNEN*GTUBE-STANDARD-ANTI-UBE-TEST-EMAIL*C.34X
 """
 
+GLOBALSCOPE = '$GLOBAL'
+if SQL_EXTENSION_ENABLED:
+    from sqlalchemy import Column
+    from sqlalchemy.types import Unicode, Integer
+    
+    class UserPref(DeclarativeBase):
+        __tablename__ = 'userpref'
+        prefid = Column(Integer, primary_key=True)
+        username = Column(Unicode(100),nullable=False)
+        preference = Column(Unicode(30),nullable=False)
+        value = Column(Unicode(100),nullable=False)
+else:
+    UserPref = None
+
 
 
 class SAPlugin(ScannerPlugin):
@@ -89,6 +102,11 @@ Tags:
                 'description': 'how often should fuglu retry the connection before giving up',
             },
 
+            'retry_sleep': {
+                'default': '1',
+                'description': 'how long should fuglu wait in seconds before retryng the connection',
+            },
+
             'scanoriginal': {
                 'default': 'True',
                 'description': "should we scan the original message as retreived from postfix or scan the current state \nin fuglu (which might have been altered by previous plugins)\nonly set this to disabled if you have a custom plugin that adds special headers to the message that will be \nused in spamassassin rules",
@@ -144,21 +162,11 @@ Tags:
                 'description': "reject message template if running in pre-queue mode",
             },
 
-            'check_sql_blacklist': {
-                'default': 'False',
-                'description': "consult spamassassins(or any other) sql blacklist for messages that are too big for spam checks\nrequires the sql extension to be enabled",
-            },
-
-            'sql_blacklist_dbconnectstring': {
-                'default': 'mysql:///localhost/spamassassin',
-                'description': "sqlalchemy db connect string",
-                'confidential': True,
-            },
-
-            'sql_blacklist_sql': {
-                'default': """SELECT value FROM userpref WHERE prefid='blacklist_from' AND username in ('$GLOBAL',concat('%',${to_domain}),${to_address})""",
-                'description': "SQL query to get the blacklist entries for a suspect\nyou may use template variables: ${from_address} ${from_domain} ${to_address} ${to_domain}",
+            'sql_blocklist_dbconnectstring': {
+                'default': '',
+                'description': "sqlalchemy db connect string, e.g. mysql:///localhost/spamassassin",
             },
+            
             'attach_suspect_tags': {
                 'default': '',
                 'description': "Suspect tags to attach as text part to message for scanning",
@@ -173,34 +181,33 @@ Tags:
 
 
     def lint(self):
-        allok = self.check_config() and self.lint_ping() and self.lint_spam() and self.lint_blacklist()
+        allok = self.check_config() and self._lint_ping() and self._lint_spam() and self._lint_blocklist()
         return allok
 
 
-    def lint_blacklist(self):
-        if not self.config.has_option(self.section, 'check_sql_blacklist') or not self.config.getboolean(self.section, 'check_sql_blacklist'):
+    def _lint_blocklist(self):
+        dbconnectionstring = self.config.get(self.section, 'sql_blocklist_dbconnectstring')
+        if not dbconnectionstring:
             return True
 
         if not SQL_EXTENSION_ENABLED:
             print("SQL Blacklist requested but SQLALCHEMY is not enabled")
             return False
-
-        session = get_session(self.config.get(self.section, 'sql_blacklist_dbconnectstring'))
+        
         suspect = Suspect('dummy@example.com', 'dummy@example.com', '/dev/null')
-        conf_sql = self.config.get(self.section, 'sql_blacklist_sql')
-        sql, params = self._replace_sql_params(suspect, conf_sql)
         try:
-            session.execute(sql, params)
-            print("Blacklist SQL Query OK")
+            self._get_sql_blocklist(suspect)
+            print("SQL UserPref blocklist query OK")
             return True
         except Exception as e:
-            print(e)
+            print("SQL UserPref blocklist query failed: %s" % str(e))
             return False
 
 
-    def lint_ping(self):
+    def _lint_ping(self):
         """ping sa"""
         retries = self.config.getint(self.section, 'retries')
+        retry_sleep = self.config.getint(self.section, 'retry_sleep')
         for i in range(0, retries):
             try:
                 self.logger.debug('Contacting spamd (Try %s of %s)' % (i + 1, retries))
@@ -230,102 +237,79 @@ Tags:
             except socket.error as e:
                 print('SPAMD socket error: %s' % str(e))
 
-            time.sleep(1)
+            time.sleep(retry_sleep)
         return False
 
 
-    def lint_spam(self):
-        spamflag, score, rules = self.safilter_symbols(GTUBE, 'test')
+    def _lint_spam(self):
+        spamflag, score, rules = self.safilter_symbols(GTUBE, 'test', fid="<no-suspect>")
         if 'GTUBE' in rules:
             print("GTUBE Has been detected correctly")
             return True
         else:
             print("SA did not detect GTUBE")
             return False
+    
+    
+    def _get_sql_blocklist(self, suspect: Suspect):
+        dbconnectionstring = self.config.get(self.section, 'sql_blocklist_dbconnectstring')
+        if not dbconnectionstring:
+            return []
+        dbsession = get_session(dbconnectionstring)
+        query = dbsession.query(UserPref)
+        query = query.filter(UserPref.preference.in_(['blacklist_from', 'blocklist_from']))
+        scopes = [
+            '$GLOBAL',
+            f'%{suspect.to_domain}',  # sa domain wide
+            f'*@{suspect.to_domain}', # roundcube sauserprefs plugin
+            suspect.to_address,
+        ]
+        query = query.filter(UserPref.username.in_(scopes))
+        results = query.all()
+        return results
+        
 
 
-    def _replace_sql_params(self, suspect, conf_sql):
-        """replace template variables in sql with parameters and set sqlalchemy parameters from suspect"""
-        values = {}
-        values['from_address'] = ':fromaddr'
-        values['to_address'] = ':toaddr'
-        values['from_domain'] = ':fromdomain'
-        values['to_domain'] = ':todomain'
-
-        template = Template(conf_sql)
-
-        sql = template.safe_substitute(values)
-
-        params = {
-            'fromaddr': suspect.from_address,
-            'toaddr': suspect.to_address,
-            'fromdomain': suspect.from_domain,
-            'todomain': suspect.to_domain,
-        }
-
-        return sql, params
-
-
-    def check_sql_blacklist(self, suspect, runtimeconfig=None):
-        """Check this message against the SQL blacklist. returns highspamaction on hit, DUNNO otherwise"""
-        #work in progress
-        if not self.config.has_option(self.section, 'check_sql_blacklist') or not self.config.getboolean(self.section, 'check_sql_blacklist'):
-            return DUNNO
-
+    def check_sql_blocklist(self, suspect, runtimeconfig=None):
+        """Check this message against the SQL blocklist. returns highspamaction on hit, DUNNO otherwise"""
         if not SQL_EXTENSION_ENABLED:
-            self.logger.error('Cannot check sql blacklist, SQLALCHEMY extension is not available')
+            self.logger.error('Cannot check sql blocklist, SQLALCHEMY extension is not available')
             return DUNNO
-
-        dbconnectionstring = self.config.get(self.section, 'sql_blacklist_dbconnectstring')
-        conf_sql = self.config.get(self.section, 'sql_blacklist_sql')
+        
         try:
-            dbsession = get_session(dbconnectionstring)
-            sql, params = self._replace_sql_params(suspect, conf_sql)
-            resultproxy = dbsession.execute(sql, params)
+            blocklistings = self._get_sql_blocklist(suspect)
         except Exception as e:
-            self.logger.error('Could not read blacklist from DB connection %s: %s' % (dbconnectionstring, str(e)))
-            suspect.debug('Blacklist check failed: %s' % e)
+            self.logger.error('%s Could not read UserPrefs: %s' % (suspect.id, str(e)))
             return DUNNO
-
-        for result in resultproxy:
-            dbvalue = result[0]  # this value might have multiple words
-            allvalues = dbvalue.split()
-            for blvalue in allvalues:
-                self.logger.debug(blvalue)
-                # build regex
-                # translate glob to regexr
-                # http://stackoverflow.com/questions/445910/create-regex-from-glob-expression
-                regexp = re.escape(blvalue).replace(r'\?', '.').replace(r'\*', '.*?')
-                self.logger.debug(regexp)
-                pattern = re.compile(regexp)
-
-                if pattern.search(suspect.from_address):
-                    self.logger.debug('%s Blacklist match : %s for sa pref %s' % (suspect.id, suspect.from_address, blvalue))
-                    confcheck = self.config
-                    if runtimeconfig is not None:
-                        confcheck = runtimeconfig
-                    configaction = string_to_actioncode(confcheck.get(self.section, 'highspamaction'), self.config)
-                    suspect.tags['spam']['SpamAssassin'] = True
-                    prependheader = self.config.get( 'main', 'prependaddedheaders')
-                    suspect.addheader("%sBlacklisted" % prependheader, blvalue)
-                    suspect.debug('Sender is Blacklisted: %s' % blvalue)
-                    if configaction is None:
-                        return DUNNO
-                    return configaction
-
+        
+        for result in blocklistings:
+            self.logger.debug('%s checking blocklist value %s' % (suspect.id, result.value))
+            # build regex
+            # translate glob to regexr
+            # http://stackoverflow.com/questions/445910/create-regex-from-glob-expression
+            regexp = re.escape(result.value).replace(r'\?', '.').replace(r'\*', '.*?')
+            self.logger.debug(regexp)
+            pattern = re.compile(regexp)
+            
+            if pattern.search(suspect.from_address):
+                self.logger.debug('%s Blocklist match : %s for sa pref %s' % (suspect.id, suspect.from_address, result.value))
+                confcheck = self.config
+                if runtimeconfig is not None:
+                    confcheck = runtimeconfig
+                configaction = string_to_actioncode(confcheck.get(self.section, 'highspamaction'), self.config)
+                highspamlevel = self.config.getfloat(self.section, 'highspamlevel')
+                self._spamreport(suspect, True, True, None, highspamlevel, enginename='SpamAssassin')
+                prependheader = self.config.get( 'main', 'prependaddedheaders')
+                suspect.add_header("%sBlacklisted" % prependheader, result.value) # deprecated
+                suspect.add_header("%sBlocklisted" % prependheader, result.value)
+                suspect.debug('Sender is Blocklisted: %s' % result.value)
+                if configaction is None:
+                    return DUNNO
+                return configaction
+        
         return DUNNO
 
 
-    def _problemcode(self):
-        retcode = string_to_actioncode(
-            self.config.get(self.section, 'problemaction'), self.config)
-        if retcode is not None:
-            return retcode
-        else:
-            # in case of invalid problem action
-            return DEFER
-
-
     def _extract_spamstatus(self, msgrep, spamheadername, suspect):
         """
         extract spamstatus and score from messages returned by spamassassin
@@ -378,7 +362,8 @@ Tags:
             stripped = True
             # send maxsize-1 to be consistent with previous implementation
             content = suspect.source_stripped_attachments(content=content, maxsize=maxsize - 1, with_mime_headers=True)
-            suspect.set_tag('SAPlugin.stripped', True)
+            suspect.set_tag('SAPlugin.stripped', True) # deprecated
+            suspect.set_tag('SpamAssassin.stripped', True)
             suspect.write_sa_temp_header('X-Fuglu-OrigSize', str(len(content_orig)))
             self.logger.info('%s stripped attachments, body size reduced from %s to %s bytes' % (
             suspect.id, len(content_orig), len(content)))
@@ -418,6 +403,9 @@ Tags:
             else:
                 content += b"\r\n" + b"\r\n".join(force_bString(extralines))
 
+            nadd_header = b'X-Fuglu-XTRA: %i\r\n' % len(extralines)
+            content = nadd_header + content
+
         return content, content_orig, stripped
 
     @staticmethod
@@ -454,9 +442,10 @@ Tags:
 
     def examine(self, suspect):
         # check if someone wants to skip sa checks
-        if suspect.get_tag('SAPlugin.skip') is True:
+        if suspect.get_tag('SAPlugin.skip') is True or suspect.get_tag('SpamAssassin.skip') is True:
             self.logger.debug('%s Skipping SA Plugin (requested by previous plugin)' % suspect.id)
-            suspect.set_tag('SAPlugin.skipreason', 'requested by previous plugin')
+            suspect.set_tag('SAPlugin.skipreason', 'requested by previous plugin') # deprecated
+            suspect.set_tag('SpamAssassin.skipreason', 'requested by previous plugin')
             return DUNNO
         
         runtimeconfig = DBConfig(self.config, suspect)
@@ -468,31 +457,34 @@ Tags:
             self.logger.info('%s Size Skip, %s > %s' % (suspect.id, suspect.size, maxsize))
             suspect.debug('Too big for spamchecks. %s > %s' % (suspect.size, maxsize))
             prependheader = self.config.get('main', 'prependaddedheaders')
-            suspect.addheader("%sSA-SKIP" % prependheader, 'Too big for spamchecks. %s > %s' % (suspect.size, maxsize))
-            suspect.set_tag('SAPlugin.skipreason', 'size skip')
-            return self.check_sql_blacklist(suspect)
+            suspect.add_header("%sSA-SKIP" % prependheader, 'Too big for spamchecks. %s > %s' % (suspect.size, maxsize))
+            suspect.set_tag('SAPlugin.skipreason', 'size skip') # deprecated
+            suspect.set_tag('SpamAssassin.skipreason', 'size skip')
+            return self.check_sql_blocklist(suspect)
         
         content, content_orig, stripped = self._get_content(suspect)
         
         forwardoriginal = self.config.getboolean(self.section, 'forwardoriginal')
         if forwardoriginal:
-            ret = self.safilter_report(content, suspect.to_address)
+            ret = self.safilter_report(content, suspect.to_address, fid=suspect.id)
             if ret is None:
                 suspect.debug('SA report Scan failed - please check error log')
-                self.logger.error('%s SA report scan FAILED' % suspect.id)
-                suspect.addheader('%sSA-SKIP' % self.config.get('main', 'prependaddedheaders'), 'SA scan failed')
-                suspect.set_tag('SAPlugin.skipreason', 'scan failed')
+                self.logger.warning('%s SA report scan FAILED' % suspect.id)
+                suspect.add_header('%sSA-SKIP' % self.config.get('main', 'prependaddedheaders'), 'SA scan failed')
+                suspect.set_tag('SAPlugin.skipreason', 'scan failed') # deprecated
+                suspect.set_tag('SpamAssassin.skipreason', 'scan failed')
                 return self._problemcode()
             isspam, spamscore, report = ret
-            suspect.tags['SAPlugin.report'] = report
+            suspect.tags['SAPlugin.report'] = report # deprecated
         
         else:
-            filtered = self.safilter(content, suspect.to_address)
+            filtered = self.safilter(content, suspect.to_address, fid=suspect.id)
             if filtered is None:
                 suspect.debug('SA Scan failed - please check error log')
                 self.logger.error('%s SA scan FAILED' % suspect.id)
-                suspect.addheader('%sSA-SKIP' % self.config.get('main', 'prependaddedheaders'), 'SA scan failed')
-                suspect.set_tag('SAPlugin.skipreason', 'scan failed')
+                suspect.add_header('%sSA-SKIP' % self.config.get('main', 'prependaddedheaders'), 'SA scan failed')
+                suspect.set_tag('SAPlugin.skipreason', 'scan failed') # deprecated
+                suspect.set_tag('SpamAssassin.skipreason', 'scan failed')
                 return self._problemcode()
             else:
                 if stripped:
@@ -528,8 +520,8 @@ Tags:
             suspect.set_source(content,att_mgr_reset=(not forwardoriginal))
             spamheadername = self.config.get(self.section, 'spamheader')
             isspam, spamscore, report = self._extract_spamstatus(newmsgrep, spamheadername, suspect)
-            suspect.tags['SAPlugin.report'] = report
-            self.logger.debug('suspect %s %s %s %s' % (suspect.id, isspam, spamscore, suspect.get_tag('SAPlugin.report')))
+            suspect.tags['SAPlugin.report'] = report # deprecated
+            self.logger.debug('suspect %s %s %s %s' % (suspect.id, isspam, spamscore, report))
         
         action = DUNNO
         message = None
@@ -538,27 +530,26 @@ Tags:
             self.logger.debug('%s Message is spam' % suspect.id)
             suspect.debug('Message is spam')
             
-            configaction = string_to_actioncode(
-                runtimeconfig.get(self.section, 'lowspamaction'), self.config)
+            configaction = string_to_actioncode(runtimeconfig.get(self.section, 'lowspamaction'), self.config)
             if configaction is not None:
                 action = configaction
             values = dict(spamscore=spamscore)
-            message = apply_template(
-                self.config.get(self.section, 'rejectmessage'), suspect, values)
+            message = apply_template(self.config.get(self.section, 'rejectmessage'), suspect, values)
         else:
             self.logger.debug('%s Message is not spam' % suspect.id)
             suspect.debug('Message is not spam')
         
-        suspect.tags['spam']['SpamAssassin'] = isspam
-        suspect.tags['highspam']['SpamAssassin'] = False
+        ishighspam = False
         if spamscore is not None:
-            suspect.tags['SAPlugin.spamscore'] = spamscore
+            suspect.tags['SAPlugin.spamscore'] = spamscore # deprecated
             highspamlevel = runtimeconfig.getfloat(self.section, 'highspamlevel')
             if spamscore >= highspamlevel:
-                suspect.tags['highspam']['SpamAssassin'] = True
+                ishighspam = True
                 configaction = string_to_actioncode(runtimeconfig.get(self.section, 'highspamaction'), self.config)
                 if configaction is not None:
                     action = configaction
+        
+        self._spamreport(suspect, isspam, ishighspam, report, spamscore, enginename='SpamAssassin')
         return action, message
 
 
@@ -579,7 +570,7 @@ Tags:
             try:
                 s.connect(sock)
             except socket.error:
-                raise Exception('Could not reach spamd using unix socket %s' % sock)
+                raise socket.error(f'Could not reach spamd using unix socket {sock}')
         else:
             host = self.config.get(self.section, 'host')
             port = self.config.getint(self.section, 'port')
@@ -587,21 +578,24 @@ Tags:
             try:
                 s = socket.create_connection((host, port), timeout)
             except socket.error:
-                raise Exception('Could not reach spamd using network (%s, %s)' % (host, port))
+                raise socket.error(f'Could not reach spamd using network ({host}, {port})')
 
         return s
 
 
-    def safilter(self, messagecontent, user):
+    def safilter(self, messagecontent, user, fid: str = ""):
         """pass content to sa, return sa-processed mail"""
         retries = self.config.getint(self.section, 'retries')
+        retry_sleep = self.config.getint(self.section, 'retry_sleep')
         peruserconfig = self.config.getboolean(self.section, 'peruserconfig')
         if self.config.getboolean(self.section, 'lowercase_user'):
             user = user.lower()
         spamsize = len(messagecontent)
-        for i in range(0, retries):
+        attempts = retries
+        while attempts:
+            attempts -= 1
             try:
-                self.logger.debug('Contacting spamd (Try %s of %s)' % (i + 1, retries))
+                self.logger.debug(f'{fid} Contacting spamd  (Try {retries - attempts} of {retries})')
                 s = self.__init_socket()
                 s.sendall(force_bString('PROCESS SPAMC/1.2'))
                 s.sendall(force_bString("\r\n"))
@@ -612,7 +606,7 @@ Tags:
                     s.sendall(force_bString("\r\n"))
                 s.sendall(force_bString("\r\n"))
                 s.sendall(force_bString(messagecontent))
-                self.logger.debug('Sent %s bytes to spamd' % spamsize)
+                self.logger.debug(f'{fid} Sent {spamsize} bytes to spamd')
                 s.shutdown(socket.SHUT_WR)
                 socketfile = s.makefile("rb")
                 line1_info = socketfile.readline()
@@ -621,36 +615,49 @@ Tags:
                 line2_contentlength = socketfile.readline()
                 line3_empty = socketfile.readline()
                 content = socketfile.read()
-                self.logger.debug('Got %s message bytes from back from spamd' % len(content))
+                self.logger.debug(f'{fid} Got {len(content)} message bytes from back from spamd')
                 answer = line1_info.strip().split()
                 if len(answer) != 3:
-                    self.logger.error("Got invalid status line from spamd: %s" % line1_info)
+                    self.logger.warning(f"{fid} Got invalid status line from spamd: {line1_info}")
+                    if attempts:
+                        time.sleep(retry_sleep)
                     continue
 
                 version, number, status = answer
                 if status != 'EX_OK':
-                    self.logger.error("Got bad status from spamd: %s" % status)
+                    self.logger.error(f"{fid} Got bad status from spamd: {status}")
+                    if attempts:
+                        time.sleep(retry_sleep)
                     continue
-
+                attempts = 0
                 return content
             except socket.timeout:
-                self.logger.error('SPAMD Socket timed out.')
+                msg = f'{fid} SPAMD Socket timed out (Try {retries - attempts} of {retries}).'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.herror as h:
-                self.logger.error('SPAMD Herror encountered : %s' % str(h))
+                msg = f'{fid} SPAMD Herror encountered (Try {retries - attempts} of {retries}): {str(h)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.gaierror as g:
-                self.logger.error('SPAMD gaierror encountered: %s' % str(g))
+                msg = f'{fid} SPAMD gaierror encountered (Try {retries - attempts} of {retries}): {str(g)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.error as e:
-                self.logger.error('SPAMD socket error: %s' % str(e))
+                msg = f'{fid} SPAMD socket error (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
+            except ConnectionRefusedError as e:
+                msg = f'{fid} SPAMD connection refused (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except Exception as e:
-                self.logger.error('SPAMD communication error: %s' % str(e))
+                msg = f'{fid} SPAMD communication error (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg, exc_info=e)
 
-            time.sleep(1)
+            if attempts:
+                time.sleep(retry_sleep)
         return None
 
 
-    def safilter_symbols(self, messagecontent, user):
+    def safilter_symbols(self, messagecontent, user, fid: str = ""):
         """Pass content to sa, return spamflag, score, rules"""
-        ret = self._safilter_content(messagecontent, user, 'SYMBOLS')
+        ret = self._safilter_content(messagecontent, user, 'SYMBOLS', fid=fid)
         if ret is None:
             return None
 
@@ -661,21 +668,24 @@ Tags:
         return status, score, rules
 
 
-    def safilter_report(self, messagecontent, user):
-        return self._safilter_content(messagecontent, user, 'REPORT')
+    def safilter_report(self, messagecontent, user, fid: str = ""):
+        return self._safilter_content(messagecontent, user, 'REPORT', fid=fid)
 
 
-    def _safilter_content(self, messagecontent, user, command):
+    def _safilter_content(self, messagecontent, user, command, fid: str = ""):
         """pass content to sa, return body"""
         assert command in ['SYMBOLS', 'REPORT', ]
         retries = self.config.getint(self.section, 'retries')
+        retry_sleep = self.config.getint(self.section, 'retry_sleep')
         peruserconfig = self.config.getboolean(self.section, 'peruserconfig')
         if self.config.getboolean(self.section, 'lowercase_user'):
             user = user.lower()
         spamsize = len(messagecontent)
-        for i in range(0, retries):
+        attempts = retries
+        while attempts:
+            attempts -= 1
             try:
-                self.logger.debug('Contacting spamd  (Try %s of %s)' % (i + 1, retries))
+                self.logger.debug(f'{fid} Contacting spamd  (Try {retries - attempts} of {retries})')
                 s = self.__init_socket()
                 s.sendall(force_bString('%s SPAMC/1.2' % command))
                 s.sendall(force_bString("\r\n"))
@@ -686,47 +696,60 @@ Tags:
                     s.sendall(force_bString("\r\n"))
                 s.sendall(force_bString("\r\n"))
                 s.sendall(force_bString(messagecontent))
-                self.logger.debug('Sent %s bytes to spamd' % spamsize)
+                self.logger.debug(f'{fid} Sent {spamsize} bytes to spamd')
                 s.shutdown(socket.SHUT_WR)
                 socketfile = s.makefile("rb")
                 line1_info = force_uString(socketfile.readline())
-                self.logger.debug(line1_info)
+                self.logger.debug(f"{fid} Answer line1: {line1_info}")
                 line2_spaminfo = force_uString(socketfile.readline())
 
                 line3 = force_uString(socketfile.readline())
                 content = socketfile.read()
                 content = content.strip()
 
-                self.logger.debug('Got %s message bytes from back from spamd' % len(content))
+                self.logger.debug(f'{fid} Got {len(content)} message bytes from back from spamd')
                 answer = line1_info.strip().split()
                 if len(answer) != 3:
-                    self.logger.error("Got invalid status line from spamd: %s" % line1_info)
+                    self.logger.warning(f"{fid} Got invalid status line from spamd: {line1_info}")
+                    if attempts:
+                        time.sleep(retry_sleep)
                     continue
 
                 version, number, status = answer
                 if status != 'EX_OK':
-                    self.logger.error("Got bad status from spamd: %s" % status)
+                    self.logger.error(f"{fid} Got bad status from spamd: %s" % status)
+                    if attempts:
+                        time.sleep(retry_sleep)
                     continue
 
-                self.logger.debug('Spamd said: %s' % line2_spaminfo)
+                self.logger.debug(f'{fid} Spamd said: {line2_spaminfo}')
                 spamword, spamstatusword, colon, score, slash, required = line2_spaminfo.split()
                 spstatus = False
                 if spamstatusword == 'True':
                     spstatus = True
-
+                attempts = 0
                 return spstatus, float(score), content
             except socket.timeout:
-                self.logger.error('SPAMD Socket timed out.')
+                msg = f'{fid} SPAMD Socket timed out (Try {retries - attempts} of {retries}).'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.herror as h:
-                self.logger.error('SPAMD Herror encountered : %s' % str(h))
+                msg = f'{fid} SPAMD Herror encountered (Try {retries - attempts} of {retries}): {str(h)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.gaierror as g:
-                self.logger.error('SPAMD gaierror encountered: %s' % str(g))
+                msg = f'{fid} SPAMD gaierror encountered (Try {retries - attempts} of {retries}): {str(g)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except socket.error as e:
-                self.logger.error('SPAMD socket error: %s' % str(e))
+                msg = f'{fid} SPAMD socket error (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
+            except ConnectionRefusedError as e:
+                msg = f'{fid} SPAMD connection refused (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg)
             except Exception as e:
-                self.logger.error('SPAMD communication error: %s' % str(e))
+                msg = f'{fid} SPAMD communication error (Try {retries - attempts} of {retries}): {str(e)}'
+                self.logger.warning(msg) if attempts else self.logger.error(msg, exc_info=e)
 
-            time.sleep(1)
+            if attempts:
+                time.sleep(retry_sleep)
         return None
 
 
@@ -897,12 +920,12 @@ Tags:
                 s.shutdown(socket.SHUT_WR)
                 socketfile = s.makefile("rb")
                 line1_info = force_uString(socketfile.readline())
-                self.logger.debug(line1_info)
+                self.logger.debug(f"{fid} Answer line1: {line1_info}")
                 line2_spaminfo = force_uString(socketfile.readline())
                 
                 answer = line1_info.strip().split()
                 if len(answer) != 3:
-                    self.logger.error("Got invalid status line from spamd: %s" % line1_info)
+                    self.logger.warning("Got invalid status line from spamd: %s" % line1_info)
                     continue
                 
                 version, number, status = answer
@@ -928,8 +951,8 @@ Tags:
             except socket.error as e:
                 self.logger.error('SPAMD socket error: %s' % str(e))
             except Exception as e:
-                self.logger.error('SPAMD communication error: %s' % str(e))
-            
+                self.logger.error('SPAMD communication error: %s' % str(e), exc_info=e)
+
             time.sleep(1)
         return None
     
