@@ -130,7 +130,7 @@ Example script:
         message = ''
         debug = lambda message: self._debug(suspect, message)
         info = lambda message: self.logger.info(message)
-        warn = lambda message: self.logger.warn(message)
+        warn = lambda message: self.logger.warning(message)
 
         def stop(action=DUNNO, message=''):
             raise Stopped(action, message)
