@@ -155,7 +155,7 @@ class ProcManager(object):
                 try:
                     task = self.tasks.get(False)
                 except EmptyQueue:
-                    self.logger.error("Queue is empty! Take a poison pill!")
+                    self.logger.warning("Queue is empty! Take a poison pill!")
                     task = None
                 if task is None:  # poison pill
                     self.logger.debug("Got poison pill")
@@ -188,8 +188,8 @@ class ProcManager(object):
             remaining_timeout = max(join_timeout - tpassed, 0.05)
             worker.join(remaining_timeout)
             if worker.is_alive():
-                self.logger.error("Could not stop worker %s (pid: %u) with given timeout of %f (%f)"
-                                  % (worker, worker.pid, join_timeout, remaining_timeout))
+                self.logger.warning("Could not stop worker %s (pid: %u) with given timeout of %f (%f)"
+                                    % (worker, worker.pid, join_timeout, remaining_timeout))
                 worker.terminate()
 
         self.logger.debug("Join message listener")
@@ -295,6 +295,7 @@ def fuglu_process_worker(queue, config, shared_state, child_to_server_messages,
     logger.debug("%s: Enter service loop..." % logtools.createPIDinfo())
 
     try:
+        logger.info(f"{logtools.createPIDinfo()}: Child worker process ready")
         while True:
             workerstate.workerstate = 'waiting for task'
             logger.debug("%s: Child process waiting for task" % logtools.createPIDinfo())
