@@ -54,8 +54,9 @@ class TrackTimings(object):
         self.realtime = 0
         self.asynctime = defaultdict(lambda: 0.0)
 
-        self.timings = []
         self.port = port if port is not None else -1
+        # name is used as hash so it's possible to add to an existing timer by same name
+        self.timingshash = {}
 
     def resettimer(self):
         """Reset time tracker"""
@@ -64,7 +65,7 @@ class TrackTimings(object):
             self.starttime = self.timetracker
             self.realtime = 0
             self.asynctime = defaultdict(lambda: 0.0)
-            self.timings = []
+            self.timingshash = {}
 
     def sum_asynctime(self, delta: tp.Union[float, int], key: tp.Optional[str] = None, logid: tp.Optional[str] = None):
         #oldval = self.asynctime[key]
@@ -82,7 +83,7 @@ class TrackTimings(object):
         else:
             self._enabletimetracker = True
             self.timetracker = time.time()
-            self.timings = []
+            self.timingshash = {}
 
     def tracktime(self, tagname, plugin=False, prepender=False, appender=False, mplugin=False):
         return self._tracktime(tagname, plugin=plugin, prepender=prepender, appender=appender, mplugin=mplugin) if self.enabletimetracker else None
@@ -116,7 +117,14 @@ class TrackTimings(object):
         difftime = newtime - self.timetracker
         self.realtime = newtime - self.starttime
         self.timetracker = newtime
-        self.timings.append((tagname, difftime, plugin, prepender, appender, mplugin))
+        hsh = (tagname, plugin, prepender, appender, mplugin)
+        existing = self.timingshash.get(hsh)
+        if existing:
+            _, e_difftime, _, _, _, _ = existing
+            difftime = e_difftime + difftime
+            self.timings_logger.debug(f"After updating existing entry for {tagname}: {e_difftime} -> {difftime}")
+
+        self.timingshash[hsh] = (tagname, difftime, plugin, prepender, appender, mplugin)
 
     def _gettime(self, plugins=None, prependers=None, appenders=None, mplugins=None):
         """
@@ -133,7 +141,7 @@ class TrackTimings(object):
 
         """
         timing_list = []
-        for timeitem in self.timings:
+        for timeitem in self.timingshash.values():
             if plugins is not None and timeitem[2] != plugins:
                 continue
             if prependers is not None and timeitem[3] != prependers:
@@ -173,18 +181,22 @@ class TrackTimings(object):
             withrealtime(bool): The real time from start to end, which might differ from total time in case of async
 
         """
-        if not self.timings:
+        if not self.timingshash:
             self.timings_logger.debug('no timings to report')
             return
 
         if end:
             self.tracktime('end')
 
+        total_asynctime = 0.0
         if withrealtime:
             self.timings_logger.info('port: %u, id: %s, real: %.6f' % (self.port, suspectid, self.realtime))
             for k, val in self.asynctime.items():
                 self.timings_logger.info(f'port: {self.port}, id: {suspectid}, async{"("+k+")" if k else ""}: {val:.6f}')
-        self.timings_logger.info('port: %u, id: %s, total: %.6f' % (self.port, suspectid, self.sumtime()))
+                total_asynctime += val
+            self.timings_logger.info('port: %u, id: %s, total(async): %.3f' % (self.port, suspectid,total_asynctime))
+
+        self.timings_logger.info('port: %u, id: %s, total: %.6f' % (self.port, suspectid, self.sumtime() - total_asynctime))
         self.timings_logger.info('port: %u, id: %s, overhead: %.3f' % (self.port, suspectid,
                                                                        self.sumtime(plugins=False,
                                                                                     prependers=False,
@@ -334,9 +346,9 @@ class SessionHandler(TrackTimings):
                 # Set fuglu spam status if wanted
                 if self.config.getboolean('main', 'spamstatusheader'):
                     if suspect.is_spam():
-                        suspect.addheader("%sSpamstatus" % prependheader, 'YES')
+                        suspect.add_header("%sSpamstatus" % prependheader, 'YES')
                     else:
-                        suspect.addheader("%sSpamstatus" % prependheader, 'NO')
+                        suspect.add_header("%sSpamstatus" % prependheader, 'NO')
 
                 # how long did it all take?
                 difftime = time.time() - starttime
@@ -345,11 +357,11 @@ class SessionHandler(TrackTimings):
                 # Debug info to mail
                 if self.config.getboolean('main', 'debuginfoheader'):
                     debuginfo = str(suspect)
-                    suspect.addheader("%sDebuginfo" % prependheader, debuginfo)
+                    suspect.add_header("%sDebuginfo" % prependheader, debuginfo)
 
                 # add suspect id for tracking
                 if self.config.getboolean('main', 'suspectidheader'):
-                    suspect.addheader('%sSuspect' % prependheader, suspect.id)
+                    suspect.add_header('%sSuspect' % prependheader, suspect.id)
                 self.tracktime("Adding-Headers")
 
                 # checks done.. print out suspect status
@@ -600,7 +612,7 @@ class SessionHandler(TrackTimings):
                     "%s : Running Plugin %s" % (suspect, plugin))
                 suspect.debug('Running plugin %s' % str(plugin))
                 starttime = time.time()
-                ans = plugin.examine(suspect)
+                ans = plugin.run_examine(suspect)
                 plugintime = time.time() - starttime
                 suspect.tags['scantimes'].append((plugin.section, plugintime))
                 message = None
